# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
from hypernets.discriminators import get_previous_trials_scores, get_percentile_score

from . import history, group_id, group_id2


def test_base():
    ts = get_previous_trials_scores(history, 0, 9, group_id)
    assert ts.shape == (5, 10)
    ts = get_previous_trials_scores(history, 0, 8, group_id)
    assert ts.shape == (6, 9)
    ts2 = get_previous_trials_scores(history, 0, 9, group_id2)
    assert ts2.shape == (1, 10)

    def get_0_100_50_percentile_score(n_step, sign=-1):
        s1 = get_percentile_score(history, n_step, group_id, 0, sign)
        s2 = get_percentile_score(history, n_step, group_id, 100, sign)
        s3 = get_percentile_score(history, n_step, group_id, 50, sign)
        return s1, s2, s3

    p1 = get_0_100_50_percentile_score(0)
    assert p1 == (0.9, 0.9, 0.9)

    p2 = get_0_100_50_percentile_score(1)
    assert p2 == (0.85, 0.8, 0.85)

    p3 = get_0_100_50_percentile_score(5)
    assert p3 == (0.45, 0.4, 0.425)

    p4 = get_0_100_50_percentile_score(9)
    assert p4 == (0.25, 0.21, 0.23)

    p5 = get_0_100_50_percentile_score(9, 1)
    assert p5 == (0.21, 0.25, 0.23)



# -*- coding:utf-8 -*-
"""

"""
from hypernets.core.trial import *
from hypernets.core.search_space import *
from hypernets.core.ops import *
from hypernets.tests import test_output_dir


class Test_TrialStore():
    def get_space(self):
        space = HyperSpace()
        with space.as_default():
            id1 = Identity(p1=Choice([1, 2]), p2=Int(1, 100))
            id2 = Identity(p3=Real(0, 1, step=0.2))(id1)
            id3 = Identity(p4=Dynamic(lambda p5: p5 * 3, p5=Choice([2, 4, 8])))(id2)
        return space

    def test_basic(self):
        store = DiskTrialStore(f'{test_output_dir}/trial_store')
        dataset_id = 'test_dataset'
        sample = self.get_space()
        sample.random_sample()

        trial = Trial(sample, 1, 0.99, 100)
        store.put(dataset_id, trial)
        store.reset()

        trial_get = store.get(dataset_id, sample)
        assert trial.trial_no == 1
        assert trial.reward == 0.99
        assert trial.elapsed == 100
        assert trial.space_sample.vectors == trial_get.space_sample.vectors

        trials = store.get_all(dataset_id, sample.signature)
        assert trials



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
from hypernets.tabular.lifelong_learning import PrequentialSplit
import numpy as np


class Test_PrequentialSplit():

    def get_train_test_index(self, split, X):
        train = []
        test = []
        for train_indices, test_indices in split.split(X):
            train.append((train_indices[0], train_indices[-1]))
            test.append((test_indices[0], test_indices[-1]))
        return train, test

    def test_strategy_preq_bls(self):
        X = np.zeros((1003, 4))

        split = PrequentialSplit(strategy=PrequentialSplit.STRATEGY_PREQ_BLS, base_size=None, n_splits=5,
                                 max_train_size=200)
        train, test = self.get_train_test_index(split, X)
        assert split.fold_size == 200
        assert train == [(0, 202), (0, 402), (0, 602), (0, 802)]
        assert test == [(203, 402), (403, 602), (603, 802), (803, 1002)]

        split = PrequentialSplit(strategy=PrequentialSplit.STRATEGY_PREQ_BLS, base_size=50, n_splits=5,
                                 max_train_size=200)
        train, test = self.get_train_test_index(split, X)
        assert split.fold_size == 190
        assert train == [(0, 242), (0, 432), (0, 622), (0, 812)]
        assert test == [(243, 432), (433, 622), (623, 812), (813, 1002)]

        split = PrequentialSplit(strategy=PrequentialSplit.STRATEGY_PREQ_BLS, base_size=50, n_splits=5,
                                 stride=3,
                                 max_train_size=200)
        train, test = self.get_train_test_index(split, X)
        assert split.fold_size == 190
        assert train == [(0, 242), (0, 812)]
        assert test == [(243, 432), (813, 1002)]

    def test_strategy_preq_slid_bls(self):
        X = np.zeros((1003, 4))

        split = PrequentialSplit(strategy=PrequentialSplit.STRATEGY_PREQ_SLID_BLS, base_size=None, n_splits=5,
                                 max_train_size=200)
        train, test = self.get_train_test_index(split, X)
        assert split.fold_size == 200
        assert train == [(3, 202), (203, 402), (403, 602), (603, 802)]
        assert test == [(203, 402), (403, 602), (603, 802), (803, 1002)]

        split = PrequentialSplit(strategy=PrequentialSplit.STRATEGY_PREQ_SLID_BLS, base_size=None, n_splits=5,
                                 max_train_size=500)
        train, test = self.get_train_test_index(split, X)
        assert split.fold_size == 200
        assert train == [(0, 202), (200, 402), (103, 602), (303, 802)]
        assert test == [(203, 402), (403, 602), (603, 802), (803, 1002)]

        split = PrequentialSplit(strategy=PrequentialSplit.STRATEGY_PREQ_SLID_BLS, base_size=50, n_splits=5,
                                 max_train_size=500)
        train, test = self.get_train_test_index(split, X)
        assert split.fold_size == 190
        assert train == [(0, 242), (190, 432), (123, 622), (313, 812)]
        assert test == [(243, 432), (433, 622), (623, 812), (813, 1002)]

        split = PrequentialSplit(strategy=PrequentialSplit.STRATEGY_PREQ_SLID_BLS, base_size=50, n_splits=5,
                                 stride=3, max_train_size=500)
        train, test = self.get_train_test_index(split, X)
        assert split.fold_size == 190
        assert train == [(0, 242), (313, 812)]
        assert test == [(243, 432), (813, 1002)]

    def test_strategy_preq_bls_gap(self):
        X = np.zeros((1003, 4))

        split = PrequentialSplit(strategy=PrequentialSplit.STRATEGY_PREQ_BLS_GAP, base_size=None, n_splits=5,
                                 max_train_size=200)
        train, test = self.get_train_test_index(split, X)
        assert split.fold_size == 200
        assert train == [(0, 202), (0, 402), (0, 602)]
        assert test == [(403, 602), (603, 802), (803, 1002)]

        split = PrequentialSplit(strategy=PrequentialSplit.STRATEGY_PREQ_BLS_GAP, base_size=50, n_splits=5,
                                 max_train_size=200)
        train, test = self.get_train_test_index(split, X)
        assert split.fold_size == 190
        assert train == [(0, 242), (0, 432), (0, 622)]
        assert test == [(433, 622), (623, 812), (813, 1002)]

        split = PrequentialSplit(strategy=PrequentialSplit.STRATEGY_PREQ_BLS_GAP, base_size=50, n_splits=5,
                                 stride=2, max_train_size=200)
        train, test = self.get_train_test_index(split, X)
        assert split.fold_size == 190
        assert train == [(0, 242), (0, 622)]
        assert test == [(433, 622), (813, 1002)]



# -*- coding:utf-8 -*-
"""

"""
import dask.dataframe as dd
import numpy as np
import pandas as pd


def prepare_dataframe():
    from hypernets.tabular.datasets import dsutils
    pdf = dsutils.load_bank()
    ddf = dd.from_pandas(pdf, npartitions=2)

    return pdf, ddf


def test_max_abs_scale():
    from sklearn import preprocessing as sk_pre
    import hypernets.tabular.dask_ex as de

    TOL = 0.00001

    pdf, ddf = prepare_dataframe()

    num_columns = [k for k, t in pdf.dtypes.items()
                   if t in (np.int32, np.int64, np.float32, np.float64)]
    pdf = pdf[num_columns]
    ddf = ddf[num_columns]

    sk_s = sk_pre.MaxAbsScaler()
    sk_r = sk_s.fit_transform(pdf)

    de_s = de.MaxAbsScaler()
    de_r = de_s.fit_transform(ddf)

    delta = (sk_s.scale_ - de_s.scale_).abs().max()
    assert delta < TOL

    delta = (sk_r - de_r.compute()).abs().max().max()
    assert delta < TOL

    delta = (sk_s.inverse_transform(sk_r) - de_s.inverse_transform(de_r).compute()) \
        .abs().max().max()
    assert delta < TOL


def test_ordinal_encoder():
    from hypernets.tabular.dask_ex import SafeOrdinalEncoder
    df1 = pd.DataFrame({"A": [1, 2, 3, 4],
                        "B": ['a', 'a', 'a', 'b']})
    df2 = pd.DataFrame({"A": [1, 2, 3, 5],
                        "B": ['a', 'b', 'z', '0']})

    ec = SafeOrdinalEncoder(dtype=np.int32)
    df = ec.fit_transform(dd.from_pandas(df1, npartitions=2)).compute()
    df_expect = pd.DataFrame({"A": [1, 2, 3, 4],
                              "B": [1, 1, 1, 2]})
    # diff = (df - df_expect).values
    # assert np.count_nonzero(diff) == 0
    assert np.where(df_expect.values == df.values, 0, 1).sum() == 0

    df = ec.transform(dd.from_pandas(df2, npartitions=1)).compute()
    df_expect = pd.DataFrame({"A": [1, 2, 3, 5],
                              "B": [1, 2, 0, 0]})
    assert np.where(df_expect.values == df.values, 0, 1).sum() == 0

    df = ec.inverse_transform(dd.from_pandas(df_expect, npartitions=1)).compute()
    df_expect = pd.DataFrame({"A": [1, 2, 3, 5],
                              "B": ['a', 'b', None, None]})
    assert np.where(df_expect.values == df.values, 0, 1).sum() == 0



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
import autosklearn.classification
import autosklearn.regression
from . import BaseEstimator
from ..column_selector import column_object


class AutoSklearnEstimator(BaseEstimator):
    def __init__(self, task, **kwargs):
        super(AutoSklearnEstimator, self).__init__(task)
        if task == 'regression':
            self.automl = autosklearn.regression.AutoSklearnRegressor(**kwargs)
        else:
            self.automl = autosklearn.classification.AutoSklearnClassifier(**kwargs)
        self.name = 'auto-sklearn'

    def train(self, X, y, X_test):
        target = '__tabular_toolbox_target__'
        X.insert(0, target, y)
        obj_cols = column_object(X)
        if len(obj_cols) > 0:
            X[obj_cols] = X[obj_cols].astype('category')
        y = X.pop(target)
        self.automl.fit(X, y)

    def predict_proba(self, X):
        obj_cols = column_object(X)
        if len(obj_cols) > 0:
            X[obj_cols] = X[obj_cols].astype('category')
        return self.automl.predict_proba(X)

    def predict(self, X):
        obj_cols = column_object(X)
        if len(obj_cols) > 0:
            X[obj_cols] = X[obj_cols].astype('category')
        return self.automl.predict(X)




# hk.search(X_train, y_train, X_test, y_test, cv=False, max_trials=3)
from hn_widget.experiment_util import StepStatus

from hypernets.experiment import ExperimentCallback
from hypernets.core.callbacks import Callback
import json
from IPython.display import display_html, HTML, display
import pickle

from hypernets.experiment.compete import SpaceSearchStep
from hypernets.utils import fs
from hypernets.core.callbacks import EarlyStoppingCallback
import time
import lightgbm as lgb
import xgboost as xgb
import catboost
from xgboost.sklearn import XGBModel
from lightgbm.sklearn import LGBMModel
from catboost.core import CatBoost

from hn_widget.widget import ExperimentProcessWidget
from hn_widget.experiment_util import EarlyStoppingStatus, EarlyStoppingConfig

MAX_IMPORTANCE_NUM = 10


def extract_importances(gbm_model):

    def get_imp(n_features):
        try:
            return gbm_model.feature_importances_
        except Exception as e:
            # print(e)
            return [0 for i in range(n_features)]

    if isinstance(gbm_model, XGBModel):
        importances_pairs = list(zip(gbm_model._Booster.feature_names, get_imp(len(gbm_model._Booster.feature_names))))
    elif isinstance(gbm_model, LGBMModel):
        if hasattr(gbm_model, 'feature_name_'):
            names = gbm_model.feature_name_
        else:
            names = [f'col_{i}' for i in range(gbm_model.feature_importances_.shape[0])]
        importances_pairs = list(zip(names, get_imp(len(names))))
    elif isinstance(gbm_model, CatBoost):
        importances_pairs = list(zip(gbm_model.feature_names_, get_imp(len(gbm_model.feature_names_))))
    else:
        importances_pairs = []

    importances = {}
    for name, imp in importances_pairs:
        importances[name] = imp

    return importances


def sort_imp(imp_dict, sort_imp_dict):
    sort_imps = []
    for k in sort_imp_dict:
        sort_imps.append({
            'name': k,
            'imp': sort_imp_dict[k]
        })

    top_features = list(map(lambda x: x['name'], sorted(sort_imps, key=lambda v: v['imp'], reverse=True)[: MAX_IMPORTANCE_NUM]))

    imps = []
    for f in top_features:
        imps.append({
            'name': f,
            'imp': imp_dict[f]
        })
    return imps


def send_action(widget_id, action_type, data):
    dom_widget = DOM_WIDGETS.get(widget_id)
    if dom_widget is None:
        raise Exception(f"widget_id: {widget_id} not exists ")
    action = {'type': action_type, 'payload': data}
    # print("----action-----")
    # print(action)
    dom_widget.value = action


class ActionType:
    EarlyStopped = 'earlyStopped'
    StepFinished = 'stepFinished'
    StepBegin = 'stepBegin'
    StepError = 'stepError'
    TrialFinished = 'trialFinished'
    ExperimentFinish = 'experimentFinish'
    ExperimentBreak = 'experimentBreak'


class JupyterHyperModelCallback(Callback):

    def __init__(self):
        super(JupyterHyperModelCallback, self).__init__()
        self.widget_id = None
        self.step_index = None
        self.max_trials = None

    def set_widget_id(self, widget_id):
        self.widget_id = widget_id

    def set_step_index(self, value):
        self.step_index = value

    def on_search_start(self, hyper_model, X, y, X_eval, y_eval, cv, num_folds, max_trials, dataset_id, trial_store,
                        **fit_kwargs):
        self.max_trials = max_trials

    def on_search_end(self, hyper_model):
        for c in hyper_model.callbacks:
            if isinstance(c, EarlyStoppingCallback):
                if c.triggered:
                    # if c.triggered_reason == EarlyStoppingCallback.REASON_TIME_LIMIT:
                    #     value = c.time_limit
                    # elif c.triggered_reason == EarlyStoppingCallback.REASON_TRIAL_LIMIT:
                    #     value = c.counter_no_improvement_trials
                    # elif c.triggered_reason == EarlyStoppingCallback.REASON_EXPECTED_REWARD:
                    #     value = c.best_reward
                    # else:
                    #     raise Exception("Unseen reason " + c.triggered_reason)
                    if c.start_time is not None:
                        elapsed_time = time.time() - c.start_time
                    else:
                        elapsed_time = None
                    ess = EarlyStoppingStatus(c.best_reward, c.best_trial_no, c.counter_no_improvement_trials, c.triggered, c.triggered_reason, elapsed_time)
                    payload = {
                        'stepIndex': self.step_index,
                        'data': ess.to_dict()
                    }
                    send_action(self.widget_id, ActionType.EarlyStopped, payload)

    def on_search_error(self, hyper_model):
        pass

    def on_build_estimator(self, hyper_model, space, estimator, trial_no):
        pass

    def on_trial_begin(self, hyper_model, space, trial_no):
        pass

    @staticmethod
    def get_space_params(space):
        params_dict = {}
        for hyper_param in space.get_assigned_params():
            # param_name = hyper_param.alias[len(list(hyper_param.references)[0].name) + 1:]
            param_name = hyper_param.alias
            param_value = hyper_param.value
            # only show number param
            # if isinstance(param_value, int) or isinstance(param_value, float):
            #     if not isinstance(param_value, bool):
            #         params_dict[param_name] = param_value
            if param_name is not None and param_value is not None:
                # params_dict[param_name.split('.')[-1]] = str(param_value)
                params_dict[param_name] = str(param_value)
        return params_dict

    def ensure_number(self, value, var_name):
        if value is None:
             raise ValueError(f"Var {var_name} can not be None.")
        else:
            if not isinstance(value, float) and not isinstance(value, int):
                raise ValueError(f"Var {var_name} = {value} not a number.")

    def on_trial_end(self, hyper_model, space, trial_no, reward, improved, elapsed):
        self.ensure_number(reward, 'reward')
        self.ensure_number(trial_no, 'trail_no')
        self.ensure_number(elapsed, 'elapsed')
        trial = None
        for t in hyper_model.history.trials:
            if t.trial_no == trial_no:
                trial = t
                break

        if trial is None:
            raise Exception(f"Trial no {trial_no} is not in history")

        model_file = trial.model_file
        with fs.open(model_file, 'rb') as input:
            model = pickle.load(input)

        cv_models = model.cv_gbm_models_
        models_json = []
        is_cv = cv_models is not None and len(cv_models) > 0
        if is_cv:
            # cv is opening
            imps = []
            for m in cv_models:
                imps.append(extract_importances(m))

            imps_avg = {}
            for k in imps[0]:
                imps_avg[k] = sum([imp.get(k, 0) for imp in imps]) / 3

            for fold, m in enumerate(cv_models):
                models_json.append({
                    'fold': fold,
                    'importances': sort_imp(extract_importances(m), imps_avg)
                })
        else:
            gbm_model = model.gbm_model
            if gbm_model is None:
                raise Exception("Both cv_models or gbm_model is None ")
            imp_dict = extract_importances(gbm_model)
            models_json.append({
                'fold': None,
                'importances': sort_imp(imp_dict, imp_dict)
            })
        early_stopping_status = None
        early_stopping_config = None
        for c in hyper_model.callbacks:
            if isinstance(c, EarlyStoppingCallback):
                early_stopping_status = EarlyStoppingStatus(c.best_reward, c.best_trial_no, c.counter_no_improvement_trials, c.triggered, c.triggered_reason, time.time() - c.start_time).to_dict()
                break
        data = {
            'stepIndex': self.step_index,
            'data': {
                "trialNo": trial_no,
                "maxTrials": self.max_trials,
                "hyperParams": self.get_space_params(space),
                "models": models_json,
                "reward": reward,
                "elapsed": elapsed,
                "is_cv": is_cv,
                "metricName": hyper_model.reward_metric,
                "earlyStopping": early_stopping_status
            }
        }
        send_action(self.widget_id, ActionType.TrialFinished, data)

    def on_trial_error(self, hyper_model, space, trial_no):
        pass

    def on_skip_trial(self, hyper_model, space, trial_no, reason, reward, improved, elapsed):
        pass


DOM_WIDGETS = {}


class JupyterWidgetExperimentCallback(ExperimentCallback):

    def __init__(self):
        self.widget_id = id(self)

    @staticmethod
    def set_up_hyper_model_callback(exp, handler):
        for c in exp.hyper_model.callbacks:
            if isinstance(c, JupyterHyperModelCallback):
                handler(c)
                break

    def experiment_start(self, exp):
        self.set_up_hyper_model_callback(exp, lambda c: c.set_widget_id(self.widget_id))
        # c.set_step_index(i)
        dom_widget = ExperimentProcessWidget(exp)
        DOM_WIDGETS[self.widget_id] = dom_widget
        display(dom_widget)
        dom_widget.initData = ''  # remove init data, if refresh the page will show nothing on the browser

    def experiment_end(self, exp, elapsed):
        send_action(self.widget_id, ActionType.ExperimentFinish, {})

    def experiment_break(self, exp, error):
        send_action(self.widget_id, ActionType.ExperimentBreak, {})

    def step_start(self, exp, step):
        from hn_widget.experiment_util import get_step_index
        from hn_widget.experiment_util import StepStatus
        step_name = step
        step_index = get_step_index(exp, step_name)
        self.set_up_hyper_model_callback(exp, lambda c: c.set_step_index(step_index))
        payload = {
            'index': step_index,
            'status': StepStatus.Process,
            'start_datetime': time.time()
        }
        send_action(self.widget_id, ActionType.StepBegin, payload)

    def step_progress(self, exp, step, progress, elapsed, eta=None):
        pass

    def step_end(self, exp, step, output, elapsed):
        from hn_widget import experiment_util
        from hn_widget.experiment_util import StepStatus
        step_name = step
        step = exp.get_step(step_name)
        # setattr(step, 'status', StepStatus.Finish)
        # todo set time setattr(step, 'status', StepStatus.Finish)
        step.done_time = time.time()  # fix done_time is none
        step_index = experiment_util.get_step_index(exp, step_name)
        d = experiment_util.extract_step(step_index, step)
        send_action(self.widget_id, ActionType.StepFinished, d)

    def step_break(self, exp, step, error):
        from hn_widget.experiment_util import get_step_index
        from hn_widget.experiment_util import StepStatus
        step_name = step
        step_index = get_step_index(exp, step_name)
        self.set_up_hyper_model_callback(exp, lambda c: c.set_step_index(step_index))
        payload = {
            'index': step_index,
            'extension': {
                'reason': str(error)
            },
            'status': StepStatus.Error
        }
        send_action(self.widget_id, ActionType.StepBegin, payload)



# -*- coding:utf-8 -*-
"""

"""

import os

import dask
import pandas as pd
import pyarrow as pa
import pyarrow.filesystem as pafs
import pyarrow.parquet as pq
from dask import dataframe as dd

try:
    import fsspec

    _has_fsspec = True
except:
    _has_fsspec = False

__all__ = ('to_parquet', 'read_parquet')

pa_ensure_filesystem = pafs._ensure_filesystem


def _check_pa_filesystem(filesystem):
    if hasattr(filesystem, '_hyn_adapted_'):
        # print('*' * 20, 'return adapted fs')
        return filesystem

    # print('-' * 20, 'call arrow _ensure_filesystem')
    return pa_ensure_filesystem(filesystem)


pafs._ensure_filesystem = _check_pa_filesystem


def _check_filesystem(filesystem):
    if _has_fsspec:
        if isinstance(filesystem, fsspec.AbstractFileSystem):
            try:
                # do this line here to keep LocalFileSystem as fsspec
                filesystem = pa.fs.PyFileSystem(pa.fs.FSSpecHandler(filesystem))
            except:
                pass

    return filesystem


def _arrow_write_parquet(df, target_path, filesystem=None, **pa_options):
    filesystem = _check_filesystem(filesystem)
    tbl = pa.Table.from_pandas(df)
    pq.write_table(tbl, target_path, filesystem=filesystem, **pa_options)

    return target_path


def to_parquet(df, path, filesystem=None, delayed=False, **kwargs_pass):
    """
    Use pyarrow to store pandas or dask dataframe into parquet file(s)

    :param df: a pandas or dask dataframe instance
    :param path: the target parquet file path if df is pandas dataframe,
        or the root directory of partitioned parquet files for dask dataframe
    :param filesystem: pyarrow FileSystem or fsspec FileSystem
    :param delayed: [dask only]
    :param kwargs_pass: options passed to pyarrow.parquet.write_table
    :return: parquet file paths tuple or delayed tasks(dask only)
    """
    assert isinstance(df, (pd.DataFrame, dd.DataFrame))

    if isinstance(df, pd.DataFrame):
        result = _arrow_write_parquet(df, path, filesystem, **kwargs_pass)
        return (result,)

    # write dask dataframe

    is_local = type(filesystem).__name__.lower().find('local') >= 0 if filesystem else True
    path_sep = os.path.sep if is_local else '/'

    path = path.rstrip(path_sep)
    filenames = ["part.%i.parquet" % (i) for i in range(df.npartitions)]
    delayed_write = dask.delayed(_arrow_write_parquet)
    parts = [
        delayed_write(d, f'{path}/{filename}', filesystem, **kwargs_pass)
        for d, filename in zip(df.to_delayed(), filenames)
    ]

    if delayed:
        return parts
    else:
        result = dask.compute(parts)
        return result


def read_parquet(path, delayed=False, **kwargs_pass):
    if delayed:
        if 'filesystem' in kwargs_pass:
            filesystem = kwargs_pass.pop('filesystem')
            assert _has_fsspec and isinstance(filesystem, fsspec.AbstractFileSystem)

            return _adapted_dask_read_parquet(path, fs=filesystem, **kwargs_pass)
        else:
            return dd.read_parquet(path, **kwargs_pass)
    else:
        if 'filesystem' in kwargs_pass and _has_fsspec:
            filesystem = kwargs_pass['filesystem']
            if isinstance(filesystem, fsspec.AbstractFileSystem):
                try:
                    # do this line here to keep LocalFileSystem as fsspec
                    filesystem = pa.fs.PyFileSystem(pa.fs.FSSpecHandler(filesystem))
                    kwargs_pass['filesystem'] = filesystem
                except:
                    pass
        return pd.read_parquet(path, **kwargs_pass)


def _adapted_dask_read_parquet(
        path,
        fs,
        columns=None,
        filters=None,
        categories=None,
        index=None,
        storage_options=None,
        engine="auto",
        gather_statistics=None,
        split_row_groups=None,
        chunksize=None,
        **kwargs, ):
    """
    Adapted form dask.dataframe.read_parquet with given filesystem.
    {}
    """.format(dd.read_parquet.__doc__)

    from dask.dataframe.io.parquet.core \
        import tokenize, natural_sort_key, process_statistics, \
        set_index_columns, NONE_LABEL, ParquetSubgraph, new_dd_object

    if isinstance(columns, str):
        df = _adapted_dask_read_parquet(
            path,
            [columns],
            filters,
            categories,
            index,
            storage_options,
            engine,
            gather_statistics,
        )
        return df[columns]

    if columns is not None:
        columns = list(columns)

    name = "read-parquet-" + tokenize(
        path,
        columns,
        filters,
        categories,
        index,
        storage_options,
        engine,
        gather_statistics,
    )

    if isinstance(engine, str):
        engine = dd.io.parquet.core.get_engine(engine)

    if hasattr(path, "name"):
        path = dd.io.parquet.core.stringify_path(path)

    # == adapt start ==
    # fs, _, paths = get_fs_token_paths(path, mode="rb", storage_options=storage_options)
    if "*" in path:
        paths = [f for f in sorted(fs.glob(path)) if not fs.isdir(f)]
    else:
        paths = [path]
    # == adapt end ==

    paths = sorted(paths, key=natural_sort_key)  # numeric rather than glob ordering

    auto_index_allowed = False
    if index is None:
        # User is allowing auto-detected index
        auto_index_allowed = True
    if index and isinstance(index, str):
        index = [index]

    # if type(fs).__name__ == 'S3FileSystem':
    #     fs = AdaptedS3FSWrapper(fs)
    meta, statistics, parts, index = engine.read_metadata(
        fs,
        paths,
        categories=categories,
        index=index,
        gather_statistics=gather_statistics,
        filters=filters,
        split_row_groups=split_row_groups,
        **kwargs,
    )

    # Parse dataset statistics from metadata (if available)
    parts, divisions, index, index_in_columns = process_statistics(
        parts, statistics, filters, index, chunksize
    )

    # Account for index and columns arguments.
    # Modify `meta` dataframe accordingly
    meta, index, columns = set_index_columns(
        meta, index, columns, index_in_columns, auto_index_allowed
    )
    if meta.index.name == NONE_LABEL:
        meta.index.name = None

    subgraph = ParquetSubgraph(name, engine, fs, meta, columns, index, parts, kwargs)

    # Set the index that was previously treated as a column
    if index_in_columns:
        meta = meta.set_index(index)
        if meta.index.name == NONE_LABEL:
            meta.index.name = None

    if len(divisions) < 2:
        # empty dataframe - just use meta
        subgraph = {(name, 0): meta}
        divisions = (None, None)

    return new_dd_object(subgraph, name, meta, divisions)


class AdaptedS3FSWrapper(pa.filesystem.S3FSWrapper):

    def walk(self, path, refresh=False):
        # """
        # Directory tree generator, like os.walk.
        #
        # Generator version of what is in s3fs, which yields a flattened list of
        # files.
        # """
        # path = _sanitize_s3(_stringify_path(path))
        # directories = set()
        # files = set()
        #
        # for key in list(self.fs._ls(path, refresh=refresh)):
        #     path = key['Key']
        #     if key['StorageClass'] == 'DIRECTORY':
        #         directories.add(path)
        #     elif key['StorageClass'] == 'BUCKET':
        #         pass
        #     else:
        #         files.add(path)
        #
        # # s3fs creates duplicate 'DIRECTORY' entries
        # files = sorted([posixpath.split(f)[1] for f in files
        #                 if f not in directories])
        # directories = sorted([posixpath.split(x)[1]
        #                       for x in directories])
        #
        # yield path, directories, files
        #
        # for directory in directories:
        #     yield from self.walk(directory, refresh=refresh)
        yield from self.fs.walk(path, refresh=refresh)



# -*- coding:utf-8 -*-
import argparse

from hypernets.dispatchers.cluster import Cluster


def main():
    parser = argparse.ArgumentParser('run HyperNets experiment in cluster.')
    parser.add_argument('--experiment', '-experiment',
                        default=None,
                        help='experiment id, default current timestamp')
    parser.add_argument('--driver-broker', '-driver-broker',
                        help='address of the driver broker'
                             ', eg: grpc://<hostname>:<broker_port> to use grpc process broker'
                             ', or just <hostname> to use ssh process')
    parser.add_argument('--driver-port', '-driver-port',
                        type=int, default=8001,
                        help='tcp port of the driver'
                             ', the executors will connect to this port with grpc'
                             ', default 8001')
    parser.add_argument('--executor-brokers', '-executor-brokers',
                        required=True,
                        help='addresses of the executor nodes, separated by comma. '
                             'eg: "grpc://<executor1_hostname>:<broker_port>,'
                             'grpc://<executor2_hostname>:<broker_port>"')
    parser.add_argument('--with-driver', '-with-driver',
                        type=int, default=1,
                        help='start driver progress or not, default 1')
    parser.add_argument('--spaces-dir', '-spaces-dir',
                        default='tmp',
                        help='driver directory to store space files, default "tmp"')
    parser.add_argument('--logs-dir', '-logs-dir',
                        default='logs',
                        help='local directory to store log files')
    parser.add_argument('--report-interval', '-report-interval',
                        type=int, default=60,
                        help='report cluster processes, default 60')
    args, argv = parser.parse_known_args()

    cluster = Cluster(args.experiment,
                      args.driver_broker,
                      args.driver_port,
                      args.with_driver,
                      args.executor_brokers.split(','),
                      args.spaces_dir,
                      args.logs_dir,
                      args.report_interval,
                      *argv)
    cluster.run()


if __name__ == '__main__':
    try:
        main()
        print('done')
    except KeyboardInterrupt as e:
        print('KeyboardInterrupt')



# -*- coding:utf-8 -*-
"""

"""
from .search_space import ModuleSpace, Bool, Choice, MultipleChoice
import itertools


class HyperInput(ModuleSpace):
    def __init__(self, space=None, name=None, **hyperparams):
        ModuleSpace.__init__(self, space, name, **hyperparams)

    def _compile(self):
        pass

    def _forward(self, inputs):
        return inputs


class Identity(ModuleSpace):
    def __init__(self, space=None, name=None, **hyperparams):
        ModuleSpace.__init__(self, space, name, **hyperparams)

    def _compile(self):
        pass

    def _forward(self, inputs):
        return inputs


class ConnectionSpace(ModuleSpace):
    def __init__(self, dynamic_fn, keep_link=False, space=None, name=None, **hyperparams):
        self.dynamic_fn = dynamic_fn
        self.keep_link = keep_link
        ModuleSpace.__init__(self, space, name, **hyperparams)

    def _on_params_ready(self):
        with self.space.as_default():
            try:
                self.space.scope.entry(self.id)
                input, output = self.dynamic_fn(self)
                if not all([input, output]) or (
                        isinstance(input, list) and isinstance(output, list) and (len(input) <= 0 or len(output) <= 0)):
                    # node removed
                    if self.keep_link:
                        # Only one input and one output are allowed
                        inputs = self.space.get_inputs(self)
                        outputs = self.space.get_outputs(self)
                        if len(inputs) > 1 or len(outputs) > 1:
                            raise ValueError("Only one input and one output are allowed.")
                        self.space.disconnect_all(self)
                        outputs[0](inputs[0])
                    else:
                        self.space.disconnect_all(self)

                elif all([input, output]):
                    self.space.reroute_to(self, input)
                    self.space.reroute_from(self, output)
                else:
                    raise ValueError('input or output is None.')
            finally:
                self.space.scope.exit()

    def connect_module_or_subgraph(self, from_module, to_module, support_subgraph=True):
        if not support_subgraph:
            assert isinstance(from_module, ModuleSpace)
            assert isinstance(to_module, ModuleSpace)
            assert self.space.is_isolated_module(from_module), f'`from_module` is not an isolated module. '
            assert self.space.is_isolated_module(from_module), f'`to_module` is not an isolated module. '
            return to_module(from_module)
        else:
            assert isinstance(from_module, (ModuleSpace, list))
            assert isinstance(to_module, (ModuleSpace, list))

            if isinstance(from_module, ModuleSpace):
                real_from = [from_module]
            else:
                assert len(from_module) > 0, f'`from_module` contains at least 1 element.'
                # If from_module is a list, take any module to get the outputs of the subgraph
                real_from = from_module[:1]

            if not self.space.is_isolated_module(real_from[0]):
                real_from = list(self.space.get_sub_graph_outputs(real_from[0]))

            if isinstance(to_module, ModuleSpace):
                real_to = [to_module]
            else:
                assert len(to_module) > 0, f'`from_module` contains at least 1 element.'
                # If to_module is a list, take any module to get the inputs of the subgraph
                real_to = to_module[:1]

            if not self.space.is_isolated_module(real_to[0]):
                real_to = list(self.space.get_sub_graph_inputs(real_to[0]))
            for m_to in real_to:
                for m_from in real_from:
                    m_to(m_from)
            return to_module


class Optional(ConnectionSpace):

    def __init__(self, module, keep_link=False, space=None, name=None, hp_opt=None):
        assert isinstance(module, ModuleSpace), f'{module} is not a valid Module. '
        self._module = module
        self.hp_opt = hp_opt if hp_opt is not None else Bool()
        ConnectionSpace.__init__(self, self.optional_fn, keep_link, space, name, hp_opt=self.hp_opt)

    def optional_fn(self, m):
        if self.hp_opt.value:
            input = self.space.get_sub_graph_inputs(self._module)
            output = self.space.get_sub_graph_outputs(self._module)
            return input, output
        else:
            return None, None


class ModuleChoice(ConnectionSpace):
    def __init__(self, module_list, keep_link=False, space=None, name=None, hp_or=None):
        assert isinstance(module_list, list), f'module_list must be a List.'
        assert len(module_list) > 0, f'module_list contains at least 1 Module.'
        assert all([isinstance(m, ModuleSpace) for m in module_list]), 'module_list can only contain Module.'

        self.hp_or = hp_or if hp_or is not None else Choice(list(range(len(module_list))))
        self._module_list = module_list
        ConnectionSpace.__init__(self, self.or_fn, keep_link, space, name, hp_or=self.hp_or)

    def or_fn(self, m):
        module = self._module_list[self.hp_or.value]
        input = self.space.get_sub_graph_inputs(module)
        output = self.space.get_sub_graph_outputs(module)
        return input, output


class Sequential(ConnectionSpace):
    def __init__(self, module_list, keep_link=False, space=None, name=None):
        assert isinstance(module_list, list), f'module_list must be a List.'
        assert len(module_list) > 0, f'module_list contains at least 1 Module.'
        assert all([isinstance(m, (ModuleSpace, list)) for m in
                    module_list]), 'module_list can only contains ModuleSpace or list.'
        self._module_list = module_list
        self.hp_lazy = Choice([0])
        ConnectionSpace.__init__(self, self.sequential_fn, keep_link, space, name, hp_lazy=self.hp_lazy)

    def sequential_fn(self, m):
        last = self._module_list[0]
        for i in range(1, len(self._module_list)):
            self.connect_module_or_subgraph(last, self._module_list[i])
            # self._module_list[i](last)
            last = self._module_list[i]
        input = self.space.get_sub_graph_inputs(last)
        output = self.space.get_sub_graph_outputs(last)
        return input, output


class Permutation(ConnectionSpace):
    def __init__(self, module_list, keep_link=False, space=None, name=None, hp_seq=None):
        assert isinstance(module_list, list), f'module_list must be a List.'
        assert len(module_list) > 1, f'module_list contains at least 2 Module.'
        assert all([isinstance(m, ModuleSpace) for m in module_list]), 'module_list can only contain Module.'
        self._module_list = module_list

        if hp_seq is None:
            p = itertools.permutations(range(len(module_list)))
            all_seq = []
            for seq in p:
                all_seq.append(seq)
            self.hp_all_seq = Choice(all_seq)
        else:
            self.hp_all_seq = hp_seq

        ConnectionSpace.__init__(self, self.permutation_fn, keep_link, space, name, hp_all_seq=self.hp_all_seq)

    def permutation_fn(self, m):
        seq = self.hp_all_seq.value
        # input = None
        last = None
        for i in seq:
            # if input is None:
            #    input = self._module_list[i]
            if last is not None:
                self.connect_module_or_subgraph(last, self._module_list[i])
                # self._module_list[i](last)
            last = self._module_list[i]
        input = self.space.get_sub_graph_inputs(last)
        output = self.space.get_sub_graph_outputs(last)
        return input, output


class Repeat(ConnectionSpace):
    def __init__(self, module_fn, keep_link=False, space=None, name=None, repeat_times=[1]):
        assert callable(module_fn), f'{module_fn} is not a callable object. '
        assert isinstance(repeat_times, list), f'`repeat_num_choices` must be a list.'
        assert all([isinstance(c, int) for c in
                    repeat_times]), f'All of the element in `repeat_num_choices` must be integer.'
        assert all(
            [c > 0 for c in repeat_times]), f'All of the element in `repeat_num_choices` must be greater than 0.'
        self.module_fn = module_fn
        self.hp_repeat_times = Choice(repeat_times)
        ConnectionSpace.__init__(self, self.repeat_fn, keep_link, space, name, hp_repeat_times=self.hp_repeat_times)

    def repeat_fn(self, m):
        repeat_times = self.hp_repeat_times.value
        module_list = [self.module_fn(step) for step in range(repeat_times)]
        last = module_list[0]
        for i in range(1, len(module_list)):
            self.connect_module_or_subgraph(last, module_list[i])
            # module_list[i](last)
            last = module_list[i]
        input = self.space.get_sub_graph_inputs(last)
        output = self.space.get_sub_graph_outputs(last)
        return input, output


class InputChoice(ConnectionSpace):
    def __init__(self, inputs, num_chosen_most=0, num_chosen_least=1, keep_link=False, space=None, name=None,
                 hp_choice=None):
        assert isinstance(inputs, list)
        connection_num = len(inputs)
        assert connection_num >= num_chosen_least, f'`inputs` contains at least {num_chosen_least} item.'
        self.inputs = inputs
        self.hp_choice = hp_choice if hp_choice is not None else MultipleChoice(list(range(connection_num)),
                                                                                num_chosen_most, num_chosen_least)
        ConnectionSpace.__init__(self, None, keep_link, space, name, hp_choice=self.hp_choice)

    def _on_params_ready(self):
        with self.space.as_default():
            for input in self.inputs:
                self.space.disconnect(input, self)
            for i, input in enumerate(self.inputs):
                if i in self.hp_choice.value:
                    self.space.connect(input, self)

            outputs = self.space.get_outputs(self)
            for output in outputs:
                self.space.reroute_to(self, output)

            self.space.disconnect_all(self)


class ConnectLooseEnd(ConnectionSpace):
    def __init__(self, inputs, keep_link=False, space=None, name=None):
        assert isinstance(inputs, list)
        self.inputs = inputs
        self.hp_lazy = Choice([0])
        ConnectionSpace.__init__(self, None, keep_link, space, name, hp_lazy=self.hp_lazy)

    def _on_params_ready(self):
        with self.space.as_default():
            for input in self.inputs:
                outputs = self.space.get_outputs(input)
                # It's not a loose end if the input has other downstream node excepts this node
                if len(set(outputs) - {self}) > 0:
                    self.space.disconnect(input, self)

            outputs = self.space.get_outputs(self)
            for output in outputs:
                self.space.reroute_to(self, output)

            self.space.disconnect_all(self)


class Reduction(ModuleSpace):
    def __init__(self, reduction_op, space=None, name=None, **hyperparams):
        ModuleSpace.__init__(self, space, name, **hyperparams)
        self.reduction_op = reduction_op

    def _compile(self):
        pv = self.param_values
        if pv.get('name') is None:
            pv['name'] = self.name
        self.compile_fn = self.reduction_op(**pv)

    def _forward(self, inputs):
        return self.compile_fn(inputs)

    def _on_params_ready(self):
        self._compile()



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
import os
import tempfile
import time

test_output_dir = tempfile.mkdtemp(prefix=time.strftime("hyn_test_%m%d%H%M_"))

os.environ['DEEPTABLES_HOME'] = test_output_dir
os.environ['HYPERNETS_HOME'] = test_output_dir



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hypernets/dispatchers/cluster/grpc/proto/spec.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='hypernets/dispatchers/cluster/grpc/proto/spec.proto',
  package='hypernets.dispatchers.cluster.grpc.proto',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n3hypernets/dispatchers/cluster/grpc/proto/spec.proto\x12(hypernets.dispatchers.cluster.grpc.proto\"\x1e\n\x0bPingMessage\x12\x0f\n\x07message\x18\x01 \x01(\t\"x\n\rSearchRequest\x12\x11\n\tsearch_id\x18\x01 \x01(\t\x12\x10\n\x08trial_no\x18\x02 \x01(\t\x12\x10\n\x08space_id\x18\x03 \x01(\t\x12\x0f\n\x07success\x18\x04 \x01(\x08\x12\x0e\n\x06reward\x18\x05 \x01(\x02\x12\x0f\n\x07message\x18\x06 \x01(\t\"\x8f\x02\n\x0eSearchResponse\x12Y\n\x04\x63ode\x18\x01 \x01(\x0e\x32K.hypernets.dispatchers.cluster.grpc.proto.SearchResponse.SearchResponseCode\x12\x11\n\tsearch_id\x18\x02 \x01(\t\x12\x10\n\x08trial_no\x18\x03 \x01(\t\x12\x10\n\x08space_id\x18\x04 \x01(\t\x12\x12\n\nspace_file\x18\x05 \x01(\t\x12\x12\n\nmodel_file\x18\x06 \x01(\t\"C\n\x12SearchResponseCode\x12\x06\n\x02OK\x10\x00\x12\x0b\n\x07WAITING\x10\x0b\x12\x0c\n\x08\x46INISHED\x10\x0c\x12\n\n\x06\x46\x41ILED\x10\x63\x32\x8a\x02\n\x0cSearchDriver\x12v\n\x04ping\x12\x35.hypernets.dispatchers.cluster.grpc.proto.PingMessage\x1a\x35.hypernets.dispatchers.cluster.grpc.proto.PingMessage\"\x00\x12\x81\x01\n\x06search\x12\x37.hypernets.dispatchers.cluster.grpc.proto.SearchRequest\x1a\x38.hypernets.dispatchers.cluster.grpc.proto.SearchResponse\"\x00(\x01\x30\x01\x62\x06proto3')
)



_SEARCHRESPONSE_SEARCHRESPONSECODE = _descriptor.EnumDescriptor(
  name='SearchResponseCode',
  full_name='hypernets.dispatchers.cluster.grpc.proto.SearchResponse.SearchResponseCode',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='OK', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='WAITING', index=1, number=11,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FINISHED', index=2, number=12,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FAILED', index=3, number=99,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=456,
  serialized_end=523,
)
_sym_db.RegisterEnumDescriptor(_SEARCHRESPONSE_SEARCHRESPONSECODE)


_PINGMESSAGE = _descriptor.Descriptor(
  name='PingMessage',
  full_name='hypernets.dispatchers.cluster.grpc.proto.PingMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='message', full_name='hypernets.dispatchers.cluster.grpc.proto.PingMessage.message', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=97,
  serialized_end=127,
)


_SEARCHREQUEST = _descriptor.Descriptor(
  name='SearchRequest',
  full_name='hypernets.dispatchers.cluster.grpc.proto.SearchRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='search_id', full_name='hypernets.dispatchers.cluster.grpc.proto.SearchRequest.search_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trial_no', full_name='hypernets.dispatchers.cluster.grpc.proto.SearchRequest.trial_no', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='space_id', full_name='hypernets.dispatchers.cluster.grpc.proto.SearchRequest.space_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='success', full_name='hypernets.dispatchers.cluster.grpc.proto.SearchRequest.success', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='reward', full_name='hypernets.dispatchers.cluster.grpc.proto.SearchRequest.reward', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='hypernets.dispatchers.cluster.grpc.proto.SearchRequest.message', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=129,
  serialized_end=249,
)


_SEARCHRESPONSE = _descriptor.Descriptor(
  name='SearchResponse',
  full_name='hypernets.dispatchers.cluster.grpc.proto.SearchResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='hypernets.dispatchers.cluster.grpc.proto.SearchResponse.code', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='search_id', full_name='hypernets.dispatchers.cluster.grpc.proto.SearchResponse.search_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trial_no', full_name='hypernets.dispatchers.cluster.grpc.proto.SearchResponse.trial_no', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='space_id', full_name='hypernets.dispatchers.cluster.grpc.proto.SearchResponse.space_id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='space_file', full_name='hypernets.dispatchers.cluster.grpc.proto.SearchResponse.space_file', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='model_file', full_name='hypernets.dispatchers.cluster.grpc.proto.SearchResponse.model_file', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SEARCHRESPONSE_SEARCHRESPONSECODE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=252,
  serialized_end=523,
)

_SEARCHRESPONSE.fields_by_name['code'].enum_type = _SEARCHRESPONSE_SEARCHRESPONSECODE
_SEARCHRESPONSE_SEARCHRESPONSECODE.containing_type = _SEARCHRESPONSE
DESCRIPTOR.message_types_by_name['PingMessage'] = _PINGMESSAGE
DESCRIPTOR.message_types_by_name['SearchRequest'] = _SEARCHREQUEST
DESCRIPTOR.message_types_by_name['SearchResponse'] = _SEARCHRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PingMessage = _reflection.GeneratedProtocolMessageType('PingMessage', (_message.Message,), {
  'DESCRIPTOR' : _PINGMESSAGE,
  '__module__' : 'hypernets.dispatchers.cluster.grpc.proto.spec_pb2'
  # @@protoc_insertion_point(class_scope:hypernets.dispatchers.cluster.grpc.proto.PingMessage)
  })
_sym_db.RegisterMessage(PingMessage)

SearchRequest = _reflection.GeneratedProtocolMessageType('SearchRequest', (_message.Message,), {
  'DESCRIPTOR' : _SEARCHREQUEST,
  '__module__' : 'hypernets.dispatchers.cluster.grpc.proto.spec_pb2'
  # @@protoc_insertion_point(class_scope:hypernets.dispatchers.cluster.grpc.proto.SearchRequest)
  })
_sym_db.RegisterMessage(SearchRequest)

SearchResponse = _reflection.GeneratedProtocolMessageType('SearchResponse', (_message.Message,), {
  'DESCRIPTOR' : _SEARCHRESPONSE,
  '__module__' : 'hypernets.dispatchers.cluster.grpc.proto.spec_pb2'
  # @@protoc_insertion_point(class_scope:hypernets.dispatchers.cluster.grpc.proto.SearchResponse)
  })
_sym_db.RegisterMessage(SearchResponse)



_SEARCHDRIVER = _descriptor.ServiceDescriptor(
  name='SearchDriver',
  full_name='hypernets.dispatchers.cluster.grpc.proto.SearchDriver',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=526,
  serialized_end=792,
  methods=[
  _descriptor.MethodDescriptor(
    name='ping',
    full_name='hypernets.dispatchers.cluster.grpc.proto.SearchDriver.ping',
    index=0,
    containing_service=None,
    input_type=_PINGMESSAGE,
    output_type=_PINGMESSAGE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='search',
    full_name='hypernets.dispatchers.cluster.grpc.proto.SearchDriver.search',
    index=1,
    containing_service=None,
    input_type=_SEARCHREQUEST,
    output_type=_SEARCHRESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_SEARCHDRIVER)

DESCRIPTOR.services_by_name['SearchDriver'] = _SEARCHDRIVER

# @@protoc_insertion_point(module_scope)



#

import gzip
import os
import queue
import time
from os.path import exists
from threading import Thread

from hypernets.dispatchers.predict.grpc.predict_client import PredictClient
from hypernets.utils import logging

logger = logging.get_logger(__name__)


class ChunkFile(object):
    def __init__(self, data_file, result_file):
        super(ChunkFile, self).__init__()

        self.data_file = data_file
        self.result_file = result_file
        self.try_count = 0

    @property
    def data_ready_tag_file(self):
        return f'{self.data_file}.done'

    @property
    def result_ready_tag_file(self):
        return f'{self.result_file}.done'

    @property
    def data_ready(self):
        return exists(self.data_file) and exists(self.data_ready_tag_file)

    @property
    def result_ready(self):
        return exists(self.result_file) and exists(self.result_ready_tag_file)


class PredictHelper(object):
    def __init__(self, servers):
        super(PredictHelper, self).__init__()
        assert isinstance(servers, (list, tuple)) and len(servers) > 0

        self.servers = [s for s in servers if len(s) > 0]
        assert len(self.servers) > 0

    def predict(self, data_file, result_file, chunk_line_limit):
        chunks = []
        q = queue.Queue()
        status = {'running': True}

        pts = [Thread(target=self.do_predict, args=[s, q, status]) for s in self.servers]
        for p in pts:
            p.start()

        mt = Thread(target=self.merge, args=[chunks, result_file], name='MergeThread')
        mt.start()

        for chunk_file_name in self.split(data_file, chunk_line_limit):
            chunk = ChunkFile(chunk_file_name, f'{chunk_file_name}.result')
            self.touch(chunk.data_ready_tag_file)
            if exists(chunk.result_ready_tag_file):
                os.remove(chunk.result_ready_tag_file)
            chunks.append(chunk)
            q.put(chunk)
            while q.qsize() >= len(self.servers):
                time.sleep(1)

        # mark for last chunk
        chunks.append(ChunkFile('', ''))

        # wait merge
        while mt.is_alive():
            time.sleep(1)

        status['running'] = False
        time.sleep(0.1)

        if logger.is_info_enabled():
            logger.info('-' * 20, 'predict done.')

        return 0

    @staticmethod
    def do_predict(server, chunk_queue, status):
        client = PredictClient(server)
        count = 0

        while status['running']:
            try:
                chunk = chunk_queue.get(block=False)
                if logger.is_info_enabled():
                    logger.info(f'[Predict] predict {chunk.data_file} started')
                count += 1
                code = client.predict(chunk.data_file, chunk.result_file)
                if code == 0:
                    PredictHelper.touch(chunk.result_ready_tag_file)
                    if logger.is_info_enabled():
                        logger.info(f'[Predict] predict {chunk.data_file} success')
                else:
                    if logger.is_info_enabled():
                        logger.info(f'[Predict] predict {chunk.data_file} failed, code={code}, try={chunk.try_count}')
                    chunk.try_count += 1
                    chunk_queue.put(chunk)
            except queue.Empty:
                time.sleep(0.1)
            except KeyboardInterrupt:
                break
        client.close()
        if logger.is_info_enabled():
            logger.info(f'[Predict] do_predict done, {count} chunks predicted.')

    @staticmethod
    def touch(file_name):
        with open(file_name, 'w'):
            pass

    @staticmethod
    def split(data_file, chunk_line_limit):
        total_line_number = 0
        chunk_index = 0

        if data_file.endswith('.gz'):
            op = gzip.open
        else:
            op = open

        with op(data_file, 'rt', encoding='utf-8') as   df:
            line = df.readline()
            while line and len(line) > 0:
                chunk_index += 1
                chunk_file_name = '%s.%04d' % (data_file, chunk_index)
                chunk_line_number = 0
                with op(chunk_file_name, 'wt', encoding='utf-8') as cf:
                    while line and len(line) > 0 and chunk_line_number < chunk_line_limit:
                        #         if logger.is_info_enabled():
                        #             logger.info(line,end='')
                        cf.write(line)
                        chunk_line_number += 1
                        line = df.readline()
                    cf.flush()

                total_line_number += chunk_line_number
                if logger.is_info_enabled():
                    logger.info(f'[Split] {chunk_file_name} is ready, lines = {chunk_line_number}.')
                yield chunk_file_name

        if logger.is_info_enabled():
            msg = f'[Split] >>> split {data_file} into {chunk_index} files, total line number is {total_line_number}.'
            logger.info(msg)
        return chunk_index

    @staticmethod
    def merge(chunks, result_file):
        total_line_number = 0
        chunk_index = 0

        def wait_chunk():
            while len(chunks) <= chunk_index:
                time.sleep(1)
            c = chunks[chunk_index]

            return len(c.result_file) > 0

        if result_file.endswith('.gz'):
            op = gzip.open
        else:
            op = open
        with op(result_file, 'wt', encoding='utf-8') as rf:
            while wait_chunk():
                chunk = chunks[chunk_index]
                while not chunk.result_ready:  # wait ready
                    time.sleep(1)
                #         if logger.is_info_enabled():
                #             logger.info(f'{chunk.result_file} is ready')

                chunk_line_number = 0
                with op(chunk.result_file, 'rt', encoding='utf-8') as cf:
                    line = cf.readline()
                    while line and len(line) > 0:
                        rf.write(line)
                        line = cf.readline()
                        chunk_line_number += 1
                if logger.is_info_enabled():
                    msg = f'[Merge] merge {chunk.result_file} into {result_file}, line number is {chunk_line_number}'
                    logger.info(msg)
                total_line_number += chunk_line_number

                chunk_index += 1
            rf.flush()

        if logger.is_info_enabled():
            msg = '[Merge] >>> all chunk is merged into  {result_file}, total line number is {total_line_number}'
            logger.info(msg)



# -*- coding:utf-8 -*-
"""

"""
import re
import time

import numpy as np
import pandas as pd
from lightgbm import LGBMRegressor, LGBMClassifier
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.impute import SimpleImputer
from sklearn.metrics import log_loss, mean_squared_error
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder, KBinsDiscretizer, OrdinalEncoder, StandardScaler, OneHotEncoder
from sklearn.utils import column_or_1d
from sklearn.utils.validation import check_is_fitted

from hypernets.tabular import column_selector
from hypernets.utils import logging, infer_task_type, const

try:
    import jieba

    _jieba_installed = True
except ImportError:
    _jieba_installed = False

logger = logging.get_logger(__name__)


def root_mean_squared_error(y_true, y_pred,
                            sample_weight=None,
                            multioutput='uniform_average', squared=True):
    return np.sqrt(
        mean_squared_error(y_true, y_pred, sample_weight=sample_weight, multioutput=multioutput, squared=squared))


def subsample(X, y, max_samples, train_samples, task, random_state=9527):
    stratify = None
    if X.shape[0] > max_samples:
        if task != 'regression':
            stratify = y
        X_train, _, y_train, _ = train_test_split(
            X, y, train_size=max_samples, shuffle=True, stratify=stratify
        )
        if task != 'regression':
            stratify = y_train

        X_train, X_test, y_train, y_test = train_test_split(
            X_train, y_train, train_size=train_samples, shuffle=True, stratify=stratify, random_state=random_state
        )
    else:
        if task != 'regression':
            stratify = y
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, train_size=0.5, shuffle=True, stratify=stratify
        )

    return X_train, X_test, y_train, y_test


class PassThroughEstimator(BaseEstimator):

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X

    def fit_transform(self, X, y=None):
        return X


class AsTypeTransformer(BaseEstimator):
    def __init__(self, *, dtype):
        assert dtype is not None
        self.dtype = dtype

        super(AsTypeTransformer, self).__init__()

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X.astype(self.dtype)

    def fit_transform(self, X, y=None):
        return self.transform(X)


# class SafeLabelEncoder(LabelEncoder):
#     def transform(self, y):
#         check_is_fitted(self, 'classes_')
#         y = column_or_1d(y, warn=True)
#
#         unseen = len(self.classes_)
#         y = np.array([np.searchsorted(self.classes_, x) if x in self.classes_ else unseen for x in y])
#         return y

class SafeLabelEncoder(LabelEncoder):
    def transform(self, y):
        check_is_fitted(self, 'classes_')
        y = column_or_1d(y, warn=True)

        unseen = len(self.classes_)
        lookup_table = dict(zip(self.classes_, list(range(0, unseen))))
        out = np.full(len(y), unseen)
        ind_id = 0
        for cell_value in y:
            if cell_value in lookup_table:
                out[ind_id] = lookup_table[cell_value]
            ind_id += 1
        return out


class MultiLabelEncoder(BaseEstimator):
    def __init__(self, columns=None):
        super(MultiLabelEncoder, self).__init__()

        self.columns = columns
        self.encoders = {}

    def fit(self, X, y=None):
        assert len(X.shape) == 2
        assert isinstance(X, pd.DataFrame) or self.columns is None

        if isinstance(X, pd.DataFrame):
            if self.columns is None:
                self.columns = X.columns.tolist()
            for col in self.columns:
                data = X.loc[:, col]
                if data.dtype == 'object':
                    data = data.astype('str')
                    # print(f'Column "{col}" has been convert to "str" type.')
                le = SafeLabelEncoder()
                le.fit(data)
                self.encoders[col] = le
        else:
            n_features = X.shape[1]
            for n in range(n_features):
                data = X[:, n]
                le = SafeLabelEncoder()
                le.fit(data)
                self.encoders[n] = le

        return self

    def transform(self, X):
        assert len(X.shape) == 2
        assert isinstance(X, pd.DataFrame) or self.columns is None

        if self.columns is not None:  # dataframe
            for col in self.columns:
                data = X.loc[:, col]
                if data.dtype == 'object':
                    data = data.astype('str')
                X.loc[:, col] = self.encoders[col].transform(data)
        else:
            n_features = X.shape[1]
            assert n_features == len(self.encoders.items())
            for n in range(n_features):
                X[:, n] = self.encoders[n].transform(X[:, n])

        return X

    def fit_transform(self, X, *args):
        assert len(X.shape) == 2
        assert isinstance(X, pd.DataFrame) or self.columns is None

        if isinstance(X, pd.DataFrame):
            if self.columns is None:
                self.columns = X.columns.tolist()
            for col in self.columns:
                data = X.loc[:, col]
                if data.dtype == 'object':
                    data = data.astype('str')
                    # print(f'Column "{col}" has been convert to "str" type.')
                le = SafeLabelEncoder()
                X.loc[:, col] = le.fit_transform(data)
                self.encoders[col] = le
        else:
            n_features = X.shape[1]
            for n in range(n_features):
                data = X[:, n]
                le = SafeLabelEncoder()
                X[:, n] = le.fit_transform(data)
                self.encoders[n] = le

        return X


class SafeOrdinalEncoder(OrdinalEncoder):
    __doc__ = r'Adapted from sklearn OrdinalEncoder\n' + OrdinalEncoder.__doc__

    def transform(self, X, y=None):
        if not isinstance(X, (pd.DataFrame, np.ndarray)):
            raise TypeError("Unexpected type {}".format(type(X)))

        def make_encoder(categories):
            unseen = len(categories)
            m = dict(zip(categories, range(unseen)))
            vf = np.vectorize(lambda x: m[x] if x in m.keys() else unseen)
            return vf

        values = X if isinstance(X, np.ndarray) else X.values
        encoders_ = [make_encoder(cat) for cat in self.categories_]
        result = [encoders_[i](values[:, i]) for i in range(values.shape[1])]

        if isinstance(X, pd.DataFrame):
            assert len(result) == len(X.columns)
            data = {c: result[i] for i, c in enumerate(X.columns)}
            result = pd.DataFrame(data, dtype=self.dtype)
        else:
            result = np.stack(result, axis=1)
            if self.dtype != result.dtype:
                result = result.astype(self.dtype)

        return result

    def inverse_transform(self, X):
        if not isinstance(X, (pd.DataFrame, np.ndarray)):
            raise TypeError("Unexpected type {}".format(type(X)))

        def make_decoder(categories, dtype):
            if dtype in (np.float32, np.float64, np.float):
                default_value = np.nan
            elif dtype in (np.int32, np.int64, np.int, np.uint32, np.uint64, np.uint):
                default_value = -1
            else:
                default_value = None
                dtype = np.object
            unseen = len(categories)
            vf = np.vectorize(lambda x: categories[x] if unseen > x >= 0 else default_value,
                              otypes=[dtype])
            return vf

        values = X if isinstance(X, np.ndarray) else X.values
        decoders_ = [make_decoder(cat, cat.dtype) for i, cat in enumerate(self.categories_)]
        result = [decoders_[i](values[:, i]) for i in range(values.shape[1])]

        if isinstance(X, pd.DataFrame):
            assert len(result) == len(X.columns)
            data = {c: result[i] for i, c in enumerate(X.columns)}
            result = pd.DataFrame(data)
        else:
            result = np.stack(result, axis=1)

        return result


class SafeOneHotEncoder(OneHotEncoder):
    def get_feature_names(self, input_features=None):
        """
        Override this method to remove non-alphanumeric chars from feature names
        """

        check_is_fitted(self)
        cats = self.categories_
        if input_features is None:
            input_features = ['x%d' % i for i in range(len(cats))]
        elif len(input_features) != len(self.categories_):
            raise ValueError(
                "input_features should have length equal to number of "
                "features ({}), got {}".format(len(self.categories_),
                                               len(input_features)))

        feature_names = []
        for i in range(len(cats)):
            names = [input_features[i] + '_' + str(idx) + '_' + re.sub('[^A-Za-z0-9_]+', '_', str(t))
                     for idx, t in enumerate(cats[i])]
            if self.drop_idx_ is not None and self.drop_idx_[i] is not None:
                names.pop(self.drop_idx_[i])
            feature_names.extend(names)

        return np.array(feature_names, dtype=object)


class LogStandardScaler(BaseEstimator):
    def __init__(self, copy=True, with_mean=True, with_std=True):
        super(LogStandardScaler, self).__init__()

        self.scaler = StandardScaler(copy=copy, with_mean=with_mean, with_std=with_std)
        self.min_values = None
        self.with_mean = with_mean
        self.with_std = with_std
        self.copy = copy

    def fit(self, X, y=None):
        self.X_min_values = np.min(X)
        self.scaler.fit(np.log(X - self.X_min_values + 1))
        return self

    def transform(self, X):
        X = np.log(np.clip(X - self.X_min_values + 1, a_min=1, a_max=None))
        X = self.scaler.transform(X)
        return X


class SkewnessKurtosisTransformer(BaseEstimator):
    def __init__(self, transform_fn=None, skew_threshold=0.5, kurtosis_threshold=0.5):
        self.columns_ = []
        self.skewness_threshold = skew_threshold
        self.kurtosis_threshold = kurtosis_threshold
        if transform_fn is None:
            transform_fn = np.log
        self.transform_fn = transform_fn

    def fit(self, X, y=None):
        assert len(X.shape) == 2
        self.columns_ = column_selector.column_skewness_kurtosis(X, skew_threshold=self.skewness_threshold,
                                                                 kurtosis_threshold=self.kurtosis_threshold)
        logger.info(f'SkewnessKurtosisTransformer - selected columns:{self.columns_}')
        return self

    def transform(self, X):
        assert len(X.shape) == 2
        if len(self.columns_) > 0:
            try:
                X[self.columns_] = self.transform_fn(X[self.columns_])
            except Exception as e:
                logger.error(e)
        return X


class FeatureSelectionTransformer(BaseEstimator):
    def __init__(self, task=None, max_train_samples=10000, max_test_samples=10000, max_cols=10000,
                 ratio_select_cols=0.1,
                 n_max_cols=100, n_min_cols=10, reserved_cols=None):
        super(FeatureSelectionTransformer, self).__init__()

        self.task = task
        if max_cols <= 0:
            max_cols = 10000
        if max_train_samples <= 0:
            max_train_samples = 10000
        if max_test_samples <= 0:
            max_test_samples = 10000

        self.max_train_samples = max_train_samples
        self.max_test_samples = max_test_samples
        self.max_cols = max_cols
        self.ratio_select_cols = ratio_select_cols
        self.n_max_cols = n_max_cols
        self.n_min_cols = n_min_cols
        self.reserved_cols = reserved_cols
        self.scores_ = {}
        self.columns_ = []

    def get_categorical_features(self, X):
        cat_cols = column_selector.column_object_category_bool(X)
        int_cols = column_selector.column_int(X)
        for c in int_cols:
            if X[c].min() >= 0 and X[c].max() < np.iinfo(np.int32).max:
                cat_cols.append(c)
        return cat_cols

    def feature_score(self, F_train, y_train, F_test, y_test):
        if self.task is None:
            self.task, _ = infer_task_type(y_train)

        if self.task == 'regression':
            model = LGBMRegressor()
            eval_metric = root_mean_squared_error
        else:
            model = LGBMClassifier()
            eval_metric = log_loss

        cat_cols = self.get_categorical_features(F_train)

        model.fit(F_train, y_train,
                  # eval_set=(F_test, y_test),
                  # early_stopping_rounds=20,
                  # verbose=0,
                  # categorical_feature=cat_cols,
                  # eval_metric=eval_metric,
                  )
        if self.task == 'regression':
            y_pred = model.predict(F_test)
        else:
            y_pred = model.predict_proba(F_test)[:, 1]

        score = eval_metric(y_test, y_pred)
        return score

    def fit(self, X, y):
        start_time = time.time()
        if self.task is None:
            self.task, _ = infer_task_type(y)
        columns = X.columns.to_list()
        logger.info(f'all columns: {columns}')
        if self.reserved_cols is not None:
            self.reserved_cols = list(set(self.reserved_cols).intersection(columns))
            logger.info(f'exclude reserved columns: {self.reserved_cols}')
            columns = list(set(columns) - set(self.reserved_cols))

        if len(columns) > self.max_cols:
            columns = np.random.choice(columns, self.max_cols, replace=False)

        if len(columns) <= 0:
            logger.warn('no columns to score')
            self.columns_ = self.reserved_cols
            self.scores_ = {}
            return self

        X_score = X[columns]
        X_train, X_test, y_train, y_test = subsample(X_score, y,
                                                     max_samples=self.max_test_samples + self.max_train_samples,
                                                     train_samples=self.max_train_samples,
                                                     task=self.task)
        if self.task != 'regression' and y_train.dtype != 'int':
            le = LabelEncoder()
            y_train = le.fit_transform(y_train)
            y_test = le.transform(y_test)

        cat_cols = column_selector.column_object_category_bool(X_train)

        if len(cat_cols) > 0:
            logger.info('ordinal encoding...')
            X_train['__datacanvas__source__'] = 'train'
            X_test['__datacanvas__source__'] = 'test'
            X_all = pd.concat([X_train, X_test], axis=0)
            oe = OrdinalEncoder()
            X_all[cat_cols] = oe.fit_transform(X_all[cat_cols]).astype('int')

            X_train = X_all[X_all['__datacanvas__source__'] == 'train']
            X_test = X_all[X_all['__datacanvas__source__'] == 'test']
            X_train.pop('__datacanvas__source__')
            X_test.pop('__datacanvas__source__')

        self.scores_ = {}

        for c in columns:
            F_train = X_train[[c]]
            F_test = X_test[[c]]
            self.scores_[c] = self.feature_score(F_train, y_train, F_test, y_test)
            logger.info(f'Feature score: {c}={self.scores_[c]}')

        sorted_scores = sorted([[col, score] for col, score in self.scores_.items()], key=lambda x: x[1])
        logger.info(f'feature scores:{sorted_scores}')
        topn = np.min([np.max([int(len(columns) * self.ratio_select_cols), np.min([len(columns), self.n_min_cols])]),
                       self.n_max_cols])
        if self.reserved_cols is not None:
            self.columns_ = self.reserved_cols
        else:
            self.columns_ = []
        self.columns_ += [s[0] for s in sorted_scores[:topn]]

        logger.info(f'selected columns:{self.columns_}')
        logger.info(f'taken {time.time() - start_time}s')

        del X_score, X_train, X_test, y_train, y_test

        return self

    def transform(self, X):
        return X[self.columns_]


class FloatOutputImputer(SimpleImputer):

    def transform(self, X):
        return super().transform(X).astype(np.float64)


class LgbmLeavesEncoder(BaseEstimator, TransformerMixin):
    def __init__(self, cat_vars, cont_vars, task, **params):
        super(LgbmLeavesEncoder, self).__init__()

        self.lgbm = None
        self.cat_vars = cat_vars
        self.cont_vars = cont_vars
        self.new_columns = []
        self.task = task
        self.lgbm_params = params

    def fit(self, X, y):
        from lightgbm import LGBMClassifier, LGBMRegressor

        X[self.cont_vars] = X[self.cont_vars].astype('float')
        X[self.cat_vars] = X[self.cat_vars].astype('int')

        logger.info(f'LightGBM task:{self.task}')
        if self.task == const.TASK_MULTICLASS:  # multiclass label
            if len(y.shape) > 1 and y.shape[1] > 1:
                num_class = y.shape[-1]
                if self.lgbm_params is None:
                    self.lgbm_params = {}
                y = y.argmax(axis=-1)
            else:
                if hasattr(y, 'unique'):
                    num_class = len(set(y.unique()))
                else:
                    num_class = len(set(y))
            self.lgbm_params['num_class'] = num_class + 1
            self.lgbm_params['n_estimators'] = int(100 / num_class) + 1

        if self.task == const.TASK_REGRESSION:
            self.lgbm = LGBMRegressor(**self.lgbm_params)
        else:
            self.lgbm = LGBMClassifier(**self.lgbm_params)
        self.lgbm.fit(X, y)
        return self

    def transform(self, X):
        X[self.cont_vars] = X[self.cont_vars].astype('float')
        X[self.cat_vars] = X[self.cat_vars].astype('int')

        leaves = self.lgbm.predict(X, pred_leaf=True, num_iteration=self.lgbm.best_iteration_)
        new_columns = [f'lgbm_leaf_{i}' for i in range(leaves.shape[1])]
        df_leaves = pd.DataFrame(leaves, columns=new_columns, index=X.index)
        result = pd.concat([X, df_leaves], axis=1)

        self.new_columns = new_columns

        return result


class CategorizeEncoder(BaseEstimator, TransformerMixin):
    def __init__(self, columns=None, remain_numeric=True):
        super(CategorizeEncoder, self).__init__()

        self.columns = columns
        self.remain_numeric = remain_numeric

        # fitted
        self.new_columns = []

    def fit(self, X, y=None):
        if self.columns is None:
            self.columns = X.columns.tolist()

        new_columns = []
        if self.remain_numeric:
            for col in self.columns:
                target_col = col + const.COLUMNNAME_POSTFIX_CATEGORIZE
                new_columns.append((target_col, 'str', X[col].nunique()))

        self.new_columns = new_columns

        return self

    def transform(self, X):
        for col in self.columns:
            if self.remain_numeric:
                target_col = col + const.COLUMNNAME_POSTFIX_CATEGORIZE
            else:
                target_col = col
            X[target_col] = X[col].astype('str')
        return X


class MultiKBinsDiscretizer(BaseEstimator, TransformerMixin):
    def __init__(self, columns=None, bins=None, strategy='quantile'):
        super(MultiKBinsDiscretizer, self).__init__()

        logger.info(f'{len(columns)} variables to discrete.')
        self.columns = columns
        self.bins = bins
        self.strategy = strategy
        self.new_columns = []
        self.encoders = {}

    def fit(self, X, y=None):
        self.new_columns = []
        if self.columns is None:
            self.columns = X.columns.tolist()
        for col in self.columns:
            new_name = col + const.COLUMNNAME_POSTFIX_DISCRETE
            n_unique = X.loc[:, col].nunique()
            n_null = X.loc[:, col].isnull().sum()
            c_bins = self.bins
            if c_bins is None or c_bins <= 0:
                c_bins = round(n_unique ** 0.25) + 1
            encoder = KBinsDiscretizer(n_bins=c_bins, encode='ordinal', strategy=self.strategy)
            self.new_columns.append((col, new_name, encoder.n_bins))
            encoder.fit(X[[col]])
            self.encoders[col] = encoder
        return self

    def transform(self, X):
        for col in self.columns:
            new_name = col + const.COLUMNNAME_POSTFIX_DISCRETE
            encoder = self.encoders[col]
            nc = encoder.transform(X[[col]]).astype(const.DATATYPE_LABEL).reshape(-1)
            X[new_name] = nc
        return X


class DataFrameWrapper(BaseEstimator, TransformerMixin):
    def __init__(self, transform, columns=None):
        super(DataFrameWrapper, self).__init__()

        self.transformer = transform
        self.columns = columns

    def fit(self, X, y=None):
        if self.columns is None:
            self.columns = X.columns.tolist()
        self.transformer.fit(X)
        return self

    def transform(self, X):
        df = pd.DataFrame(self.transformer.transform(X))
        df.columns = self.columns
        return df


class GaussRankScaler(BaseEstimator):
    def __init__(self):
        super(GaussRankScaler, self).__init__()

        self.epsilon = 0.001
        self.lower = -1 + self.epsilon
        self.upper = 1 - self.epsilon
        self.range = self.upper - self.lower
        self.divider = None

    def fit_transform(self, X, y=None):
        from scipy.special import erfinv
        i = np.argsort(X, axis=0)
        j = np.argsort(i, axis=0)

        assert (j.min() == 0).all()
        assert (j.max() == len(j) - 1).all()

        j_range = len(j) - 1
        self.divider = j_range / self.range

        transformed = j / self.divider
        transformed = transformed - self.upper
        transformed = erfinv(transformed)

        return transformed


class VarLenFeatureEncoder:
    def __init__(self, sep='|'):
        super(VarLenFeatureEncoder, self).__init__()

        self.sep = sep
        self.encoder: SafeLabelEncoder = None
        self._max_element_length = 0

    def fit(self, X: pd.Series):
        self._max_element_length = 0  # reset
        if not isinstance(X, pd.Series):
            X = pd.Series(X)
        key_set = set()
        # flat map
        for keys in X.map(lambda _: _.split(self.sep)):
            if len(keys) > self._max_element_length:
                self._max_element_length = len(keys)
            key_set.update(keys)
        key_set = list(key_set)
        key_set.sort()

        lb = SafeLabelEncoder()  # fix unseen values
        lb.fit(np.array(key_set))
        self.encoder = lb
        return self

    def transform(self, X: pd.Series):
        if self.encoder is None:
            raise RuntimeError("Not fit yet .")

        from tensorflow.python.keras.preprocessing.sequence import pad_sequences

        if not isinstance(X, pd.Series):
            X = pd.Series(X)
        # Notice : input value 0 is a special "padding",so we do not use 0 to encode valid feature for sequence input
        data = X.map(lambda _: (self.encoder.transform(_.split(self.sep)) + 1).tolist())

        transformed = pad_sequences(data, maxlen=self._max_element_length, padding='post',
                                    truncating='post').tolist()  # cut last elements
        return transformed

    @property
    def n_classes(self):
        return len(self.encoder.classes_)

    @property
    def max_element_length(self):
        return self._max_element_length


class MultiVarLenFeatureEncoder(BaseEstimator, TransformerMixin):
    def __init__(self, features):
        super(MultiVarLenFeatureEncoder, self).__init__()

        self.features = features

        # fitted
        self.encoders_ = {}  # feature name -> VarLenFeatureEncoder
        self.max_length_ = {}  # feature name -> max length

    def fit(self, X, y=None):
        encoders = {feature[0]: VarLenFeatureEncoder(feature[1]) for feature in self.features}
        max_length = {}

        for k, v in encoders.items():
            v.fit(X[k])
            max_length[k] = v.max_element_length

        self.encoders_ = encoders
        self.max_length_ = max_length

        return self

    def transform(self, X):
        for k, v in self.encoders_.items():
            X[k] = v.transform(X[k])
        return X


class LocalizedTfidfVectorizer(TfidfVectorizer):
    def decode(self, doc):
        doc = super().decode(doc)

        if _jieba_installed and self._exist_chinese(doc):
            doc = ' '.join(jieba.cut(doc))

        return doc

    @staticmethod
    def _exist_chinese(s):
        if isinstance(s, str):
            for ch in s:
                if u'\u4e00' <= ch <= u'\u9fff':
                    return True

        return False


class TfidfEncoder(BaseEstimator, TransformerMixin):
    def __init__(self, columns=None, flatten=False, **kwargs):
        assert columns is None or isinstance(columns, (str, list, tuple))
        if isinstance(columns, str):
            columns = [columns]

        super(TfidfEncoder, self).__init__()

        self.columns = columns
        self.flatten = flatten
        self.encoder_kwargs = kwargs.copy()

        # fitted
        self.encoders_ = None

    def fit(self, X, y=None):
        assert isinstance(X, (np.ndarray, pd.DataFrame)) and len(X.shape) == 2

        if self.columns is None:
            if isinstance(X, pd.DataFrame):
                columns = column_selector.column_object(X)
            else:
                columns = range(X.shape[1])
        else:
            columns = self.columns

        encoders = {}
        for c in columns:
            encoder = LocalizedTfidfVectorizer(**self.encoder_kwargs)
            Xc = X[c] if isinstance(X, pd.DataFrame) else X[:, c]
            encoders[c] = encoder.fit(Xc, y)

        self.encoders_ = encoders

        return self

    def transform(self, X, y=None):
        assert self.encoders_ is not None
        assert isinstance(X, (np.ndarray, pd.DataFrame)) and len(X.shape) == 2

        if isinstance(X, pd.DataFrame):
            X = X.copy()
            if self.flatten:
                dfs = [X]
                for c, encoder in self.encoders_.items():
                    t = encoder.transform(X[c]).toarray()
                    dfs.append(pd.DataFrame(t, index=X.index, columns=[f'{c}_tfidf_{i}' for i in range(t.shape[1])]))
                    X.pop(c)
                X = pd.concat(dfs, axis=1)
            else:
                for c, encoder in self.encoders_.items():
                    t = encoder.transform(X[c]).toarray()
                    X[c] = t.tolist()
        else:
            r = []
            tolist = None if self.flatten else np.vectorize(self._to_array, otypes=[np.object], signature='(m)->()')
            for i in range(X.shape[1]):
                Xi = X[:, i]
                if i in self.encoders_.keys():
                    encoder = self.encoders_[i]
                    t = encoder.transform(Xi).toarray()
                    if tolist is not None:
                        t = tolist(t).reshape((-1, 1))
                    r.append(t)
                else:
                    r.append(Xi)
            X = np.hstack(r)

        return X

    @staticmethod
    def _to_list(x):
        return x.tolist()

    @staticmethod
    def _to_array(x):
        return x


class DatetimeEncoder(BaseEstimator, TransformerMixin):
    all_items = ['year', 'month', 'day', 'hour', 'minute', 'second',
                 'week', 'weekday', 'dayofyear',
                 'timestamp']
    all_items = {k: k for k in all_items}
    all_items['timestamp'] = lambda t: time.mktime(t.timetuple())

    default_include = ['month', 'day', 'hour', 'minute',
                       'week', 'weekday', 'dayofyear']

    def __init__(self, columns=None, include=None, exclude=None, extra=None, drop_constants=True):
        assert columns is None or isinstance(columns, (str, list, tuple))
        assert include is None or isinstance(include, (str, list, tuple))
        assert exclude is None or isinstance(exclude, (str, list, tuple))
        assert extra is None or isinstance(extra, (tuple, list))
        if extra is not None:
            assert all(len(x) == 2 and isinstance(x[0], str)
                       and (x[1] is None or isinstance(x[1], str) or callable(x[1]))
                       for x in extra)

        if isinstance(columns, str):
            columns = [columns]

        if include is None:
            to_extract = self.default_include
        elif isinstance(include, str):
            to_extract = [include]
        else:
            to_extract = include

        if isinstance(exclude, str):
            exclude = [exclude]
        if exclude is not None:
            to_extract = [i for i in to_extract if i not in exclude]

        assert all(i in self.all_items for i in to_extract)

        to_extract = {k: self.all_items[k] for k in to_extract}
        if isinstance(extra, (tuple, list)):
            for k, c in extra:
                to_extract[k] = c

        super(DatetimeEncoder, self).__init__()

        self.columns = columns
        self.include = include
        self.exclude = exclude
        self.extra = extra
        self.drop_constants = drop_constants

        self.extract_ = to_extract

    def fit(self, X, y=None):
        if self.columns is None:
            if not isinstance(X, pd.DataFrame):
                X = pd.DataFrame(X)
            self.columns = column_selector.column_all_datetime(X)

        return self

    def transform(self, X, y=None):
        if len(self.columns) > 0:
            if isinstance(X, pd.DataFrame):
                X = X.copy()
                input_df = True
            else:
                X = pd.DataFrame(X)
                input_df = False
            dfs = [df for c in self.columns for df in self.transform_column(X[c])]
            X.drop(columns=self.columns, inplace=True)
            if len(dfs) > 0:
                dfs.insert(0, X)
                X = pd.concat(dfs, axis=1)
            if not input_df:
                X = X.values

        return X

    def transform_column(self, Xc):
        assert getattr(Xc, 'dt', None) is not None
        dfs = []

        for k, c in self.extract_.items():
            if c is None:
                c = k
            if isinstance(c, str):
                t = getattr(Xc.dt, c)
            else:
                t = Xc.apply(c)
            t.name = f'{Xc.name}_{k}'
            dfs.append(t)

        if self.drop_constants:
            dfs = [t for t in dfs if t.nunique() > 1]

        return dfs



# -*- coding:utf-8 -*-
"""

"""
import math
from functools import partial

import dask
import dask.array as da
import dask.dataframe as dd
import numpy as np
import pandas as pd
from sklearn import inspection as sk_inspect, metrics as sk_metrics
from sklearn import model_selection as sk_sel, utils as sk_utils
from sklearn.utils.multiclass import type_of_target

from hypernets.utils import logging

try:
    import dask_ml.preprocessing as dm_pre
    import dask_ml.model_selection as dm_sel

    from dask_ml.impute import SimpleImputer
    from dask_ml.compose import ColumnTransformer
    from dask_ml.preprocessing import \
        LabelEncoder, OneHotEncoder, OrdinalEncoder, \
        StandardScaler, MinMaxScaler, RobustScaler

    from ._transformers import \
        MultiLabelEncoder, SafeOneHotEncoder, TruncatedSVD, \
        MaxAbsScaler, SafeOrdinalEncoder, DataInterceptEncoder, \
        CallableAdapterEncoder, DataCacher, CacheCleaner, \
        LgbmLeavesEncoder, CategorizeEncoder, MultiKBinsDiscretizer, \
        LocalizedTfidfVectorizer, \
        MultiVarLenFeatureEncoder, DataFrameWrapper

    from ..sklearn_ex import PassThroughEstimator

    dask_ml_available = True
except ImportError:
    # Not found dask_ml
    dask_ml_available = False

logger = logging.get_logger(__name__)

compute = dask.compute


def default_client():
    try:
        from dask.distributed import default_client as dask_default_client
        client = dask_default_client()
    except ValueError:
        client = None
    return client


def dask_enabled():
    return default_client() is not None


def is_local_dask():
    client = default_client()
    return type(client.cluster).__name__.lower().find('local') >= 0 if client is not None else False


def dask_worker_count():
    client = default_client()
    return len(client.ncores()) if client else 0


def is_dask_dataframe(X):
    return isinstance(X, dd.DataFrame)


def is_dask_series(X):
    return isinstance(X, dd.Series)


def is_dask_dataframe_or_series(X):
    return isinstance(X, (dd.DataFrame, dd.Series))


def is_dask_array(X):
    return isinstance(X, da.Array)


def is_dask_object(X):
    return isinstance(X, (da.Array, dd.DataFrame, dd.Series))


def exist_dask_object(*args):
    for a in args:
        if isinstance(a, (da.Array, dd.DataFrame, dd.Series)):
            return True
        if isinstance(a, (tuple, list, set)):
            return exist_dask_object(*a)
    return False


def exist_dask_dataframe(*args):
    for a in args:
        if isinstance(a, dd.DataFrame):
            return True
        if isinstance(a, (tuple, list, set)):
            return exist_dask_dataframe(*a)
    return False


def exist_dask_array(*args):
    for a in args:
        if isinstance(a, da.Array):
            return True
        if isinstance(a, (tuple, list, set)):
            return exist_dask_array(*a)
    return False


def to_dask_type(X):
    if isinstance(X, np.ndarray):
        worker_count = dask_worker_count()
        chunk_size = math.ceil(X.shape[0] / worker_count) if worker_count > 0 else X.shape[0]
        X = da.from_array(X, chunks=chunk_size)
    elif isinstance(X, (pd.DataFrame, pd.Series)):
        worker_count = dask_worker_count()
        partition_count = worker_count if worker_count > 0 else 1
        X = dd.from_pandas(X, npartitions=partition_count)

    return X


def _reset_part_index(df, start):
    new_index = pd.RangeIndex.from_range(range(start, start + df.shape[0]))
    df.index = new_index
    return df


def reset_index(X):
    assert isinstance(X, (pd.DataFrame, dd.DataFrame))

    if is_dask_dataframe(X):
        part_rows = X.map_partitions(lambda df: pd.DataFrame({'rows': [df.shape[0]]}),
                                     meta={'rows': 'int64'},
                                     ).compute()['rows'].tolist()
        assert len(part_rows) == X.npartitions

        divisions = [0]
        n = 0
        for i in part_rows:
            n += i
            divisions.append(n)
        divisions[-1] = divisions[-1] - 1

        delayed_reset_part_index = dask.delayed(_reset_part_index)
        parts = [delayed_reset_part_index(part, start) for part, start in zip(X.to_delayed(), divisions[0:-1])]
        X_new = dd.from_delayed(parts, divisions=divisions, meta=X.dtypes.to_dict())
        return X_new
    else:
        return X.reset_index(drop=True)


def make_chunk_size_known(a):
    assert is_dask_array(a)

    chunks = a.chunks
    if any(np.nan in d for d in chunks):
        if logger.is_debug_enabled():
            logger.debug(f'call extracted array compute_chunk_sizes, shape: {a.shape}')
        a = a.compute_chunk_sizes()
    return a


def make_divisions_known(X):
    assert is_dask_object(X)

    if is_dask_dataframe(X):
        if not X.known_divisions:
            columns = X.columns.tolist()
            X = X.reset_index()
            new_columns = X.columns.tolist()
            index_name = set(new_columns) - set(columns)
            X = X.set_index(list(index_name)[0] if index_name else 'index')
            assert X.known_divisions
    elif is_dask_series(X):
        if not X.known_divisions:
            X = make_divisions_known(X.to_frame())[X.name]
    else:  # dask array
        X = make_chunk_size_known(X)

    return X


def hstack_array(arrs):
    if all([a.ndim == 1 for a in arrs]):
        rows = compute(arrs[0].shape)[0][0]
        arrs = [a.reshape(rows, 1) if a.ndim == 1 else a for a in arrs]
    return stack_array(arrs, axis=1)


def vstack_array(arrs):
    return stack_array(arrs, axis=0)


def stack_array(arrs, axis=0):
    assert axis in (0, 1)
    ndims = set([len(a.shape) for a in arrs])
    if len(ndims) > 1:
        assert ndims == {1, 2}
        assert all([len(a.shape) == 1 or a.shape[1] == 1 for a in arrs])
        arrs = [a.reshape(compute(a.shape[0])[0], 1) if len(a.shape) == 1 else a for a in arrs]
    axis = min(axis, min([len(a.shape) for a in arrs]) - 1)
    assert axis >= 0

    if exist_dask_object(*arrs):
        arrs = [a.values if is_dask_dataframe_or_series(a) else a for a in map(to_dask_type, arrs)]
        if len(arrs) > 1:
            arrs = [make_chunk_size_known(a) for a in arrs]
        return da.concatenate(arrs, axis=axis)
    else:
        return np.concatenate(arrs, axis=axis)


def array_to_df(arrs, columns=None, meta=None):
    meta_df = None
    if isinstance(meta, (dd.DataFrame, pd.DataFrame)):
        meta_df = meta
        if columns is None:
            columns = meta_df.columns
        # meta = dd.utils.make_meta(meta_df.dtypes.to_dict())
        if isinstance(meta, dd.DataFrame):
            meta = meta.head(0)
    elif isinstance(meta, (dd.Series, pd.Series)):
        meta_df = meta
        if columns is None:
            columns = meta_df.name
        meta = None

    df = dd.from_dask_array(arrs, columns=columns, meta=meta)

    if isinstance(meta_df, (dd.DataFrame, pd.DataFrame)):
        dtypes_src = meta_df.dtypes
        dtypes_dst = df.dtypes
        for col in meta_df.columns:
            if dtypes_src[col] != dtypes_dst[col]:
                df[col] = df[col].astype(dtypes_src[col])

    return df


def concat_df(dfs, axis=0, repartition=False, **kwargs):
    if exist_dask_object(*dfs):
        dfs = [dd.from_dask_array(v) if is_dask_array(v) else v for v in dfs]

        if all([isinstance(df, (dd.Series, pd.Series)) for df in dfs]):
            values = vstack_array([df.values for df in dfs])
            df = dd.from_dask_array(values, columns=dfs[0].name)
            assert isinstance(df, dd.Series)
            return df

        if axis == 0:
            values = [df[dfs[0].columns].to_dask_array(lengths=True) for df in dfs]
            df = array_to_df(vstack_array(values), meta=dfs[0])
        else:
            dfs = [make_divisions_known(df) for df in dfs]
            df = dd.concat(dfs, axis=axis, **kwargs)

        if is_dask_series(dfs[0]) and df.name is None and dfs[0].name is not None:
            df.name = dfs[0].name
        if repartition:
            df = df.repartition(npartitions=dfs[0].npartitions)
    else:
        df = pd.concat(dfs, axis=axis, **kwargs)

    return df


def train_test_split(*data, shuffle=True, random_state=None, **kwargs):
    if exist_dask_dataframe(*data):
        if len(data) > 1:
            data = [make_divisions_known(to_dask_type(x)) for x in data]
            head = data[0]
            for i in range(1, len(data)):
                if data[i].divisions != head.divisions:
                    print('-' * 10, f'repartition {i} from {data[i].divisions} to {head.divisions}')
                    data[i] = data[i].repartition(divisions=head.divisions)

        result = dm_sel.train_test_split(*data, shuffle=shuffle, random_state=random_state, **kwargs)
    else:
        result = sk_sel.train_test_split(*data, shuffle=shuffle, random_state=random_state, **kwargs)

    return result


def fix_binary_predict_proba_result(proba):
    if is_dask_object(proba):
        if proba.ndim == 1:
            proba = make_chunk_size_known(proba)
            proba = proba.reshape((proba.size, 1))
        if proba.shape[1] == 1:
            proba = hstack_array([1 - proba, proba])
    else:
        if proba.ndim == 1:
            proba = np.vstack([1 - proba, proba]).T
        elif proba.shape[1] == 1:
            proba = np.hstack([1 - proba, proba])

    return proba


def wrap_for_local_scorer(estimator, target_type):
    def _call_and_compute(fn_call, fn_fix, *args, **kwargs):
        r = fn_call(*args, **kwargs)
        if is_dask_object(r):
            r = r.compute()
            if callable(fn_fix):
                r = fn_fix(r)
        return r

    if hasattr(estimator, 'predict_proba'):
        orig_predict_proba = estimator.predict_proba
        fix = fix_binary_predict_proba_result if target_type == 'binary' else None
        setattr(estimator, '_orig_predict_proba', orig_predict_proba)
        setattr(estimator, 'predict_proba', partial(_call_and_compute, orig_predict_proba, fix))

    if hasattr(estimator, 'predict'):
        orig_predict = estimator.predict
        setattr(estimator, '_orig_predict', orig_predict)
        setattr(estimator, 'predict', partial(_call_and_compute, orig_predict, None))

    return estimator


def compute_and_call(fn_call, *args, **kwargs):
    if logger.is_debug_enabled():
        logger.debug(f'[compute_and_call] compute {len(args)} object')

    args = compute(*args, traverse=False)
    for k, v in kwargs.items():
        if exist_dask_object(v):
            kwargs[k] = compute(v, traverse=True)[0]

    if logger.is_debug_enabled():
        logger.debug(f'[compute_and_call] call {fn_call.__name__}')
    # kwargs = {k: compute(v) if is_dask_array(v) else v for k, v in kwargs.items()}
    r = fn_call(*args, **kwargs)

    if logger.is_debug_enabled():
        logger.debug('[compute_and_call] to dask type')
    r = to_dask_type(r)

    if logger.is_debug_enabled():
        logger.debug('[compute_and_call] done')
    return r


def call_and_compute(fn_call, optimize_graph, *args, **kwargs):
    if logger.is_debug_enabled():
        logger.debug(f'[call_and_compute] call {fn_call.__name__}')
    r = fn_call(*args, **kwargs)

    if is_dask_object(r):
        if logger.is_debug_enabled():
            logger.debug('[call_and_compute] to local type')
        r = compute(r, traverse=False)[0]
    elif isinstance(r, (tuple, list)) and any(map(is_dask_object, r)):
        if logger.is_debug_enabled():
            logger.debug('[call_and_compute] to local type')
        # r = compute(*r, traverse=False, optimize_graph=optimize_graph)
        r = [x.compute() if is_dask_object(x) else x for x in r]

    if logger.is_debug_enabled():
        logger.debug('[call_and_compute] done')

    return r


def wrap_local_estimator(estimator):
    for fn_name in ('fit', 'fit_cross_validation', 'predict', 'predict_proba'):
        fn_name_original = f'_wrapped_{fn_name}_by_wle'
        if hasattr(estimator, fn_name) and not hasattr(estimator, fn_name_original):
            fn = getattr(estimator, fn_name)
            assert callable(fn)
            setattr(estimator, fn_name_original, fn)
            setattr(estimator, fn_name, partial(compute_and_call, fn))

    return estimator


def permutation_importance(estimator, X, y, *args, scoring=None, n_repeats=5,
                           n_jobs=None, random_state=None):
    if not is_dask_dataframe(X):
        return sk_inspect.permutation_importance(estimator, X, y, *args,
                                                 scoring=scoring,
                                                 n_repeats=n_repeats,
                                                 n_jobs=n_jobs,
                                                 random_state=random_state)
    random_state = sk_utils.check_random_state(random_state)

    def shuffle_partition(df, col_idx):
        shuffling_idx = np.arange(df.shape[0])
        random_state.shuffle(shuffling_idx)
        col = df.iloc[shuffling_idx, col_idx]
        col.index = df.index
        df.iloc[:, col_idx] = col
        return df

    if is_dask_object(y):
        y = y.compute()

    scorer = sk_metrics.check_scoring(wrap_for_local_scorer(estimator, type_of_target(y)), scoring)
    baseline_score = scorer(estimator, X, y)
    scores = []

    for c in range(X.shape[1]):
        col_scores = []
        for i in range(n_repeats):
            X_permuted = X.copy().map_partitions(shuffle_partition, c)
            col_scores.append(scorer(estimator, X_permuted, y))
        if logger.is_debug_enabled():
            logger.debug(f'permuted scores [{X.columns[c]}]: {col_scores}')
        scores.append(col_scores)

    importances = baseline_score - np.array(scores)
    return sk_utils.Bunch(importances_mean=np.mean(importances, axis=1),
                          importances_std=np.std(importances, axis=1),
                          importances=importances)


@sk_utils._deprecate_positional_args
def compute_class_weight(class_weight, *, classes, y):
    # f"""{sk_utils.class_weight.compute_class_weight.__doc__}"""

    if not is_dask_object(y):
        return sk_utils.class_weight.compute_class_weight(class_weight, classes=classes, y=y)

    y = make_chunk_size_known(y)
    if set(compute(da.unique(y))[0]) - set(classes):
        raise ValueError("classes should include all valid labels that can be in y")

    if class_weight == 'balanced':
        # Find the weight of each class as present in y.
        le = dm_pre.LabelEncoder()
        y_ind = le.fit_transform(y)
        # if not all(np.in1d(classes, le.classes_)):
        #     raise ValueError("classes should have valid labels that are in y")
        # recip_freq = len(y) / (len(le.classes_) *
        #                        np.bincount(y_ind).astype(np.float64))
        # weight = recip_freq[le.transform(classes)]
        y_shape, y_ind_bincount, le_classes_ = compute(y.shape, da.bincount(y_ind), le.classes_)
        if not all(np.in1d(classes, le_classes_)):
            raise ValueError("classes should have valid labels that are in y")
        recip_freq = y_shape[0] / (len(le_classes_) * y_ind_bincount.astype(np.float64))
        weight = recip_freq[np.searchsorted(le_classes_, classes)]
    else:
        raise ValueError("Only class_weight == 'balanced' is supported.")

    return weight


def _compute_chunk_sample_weight(y, classes, classes_weights):
    t = np.ones(y.shape[0])
    for i, c in enumerate(classes):
        t[y == c] *= classes_weights[i]
    return t


def compute_sample_weight(y):
    assert len(y.shape) == 1 or (len(y.shape) == 2 and y.shape[1] == 1)

    if is_dask_dataframe_or_series(y):
        y = y.values

    unique = compute(da.unique(y))[0] if is_dask_object(y) else np.unique(y)
    cw = list(compute_class_weight('balanced', classes=unique, y=y))

    if is_dask_object(y):
        sample_weight = y.map_blocks(_compute_chunk_sample_weight, unique, cw, dtype=np.float64)
    else:
        sample_weight = _compute_chunk_sample_weight(y, unique, cw)

    return sample_weight



from hypernets.conf import configure, Configurable, String, Int, Float, Enum


@configure()
class DispatchCfg(Configurable):
    experiment = String(help='experiment id',
                        ).tag(config=True)
    work_dir = String(help='storage directory path to store running data.'
                      ).tag(config=True)
    backend = Enum(['standalone', 'dask', 'cluster', None],
                   default_value=None,
                   help='dispatcher backend'
                   ).tag(config=True)
    trial_retry_limit = Int(1000, min=1,
                            help='maximum retry number to run trial.'
                            ).tag(config=True)

    cluster_driver = String(help='driver address, used if backend="cluster"'
                            ).tag(config=True)
    cluster_role = Enum(['driver', 'executor'],
                        help='node role, used if backend="cluster"'
                        ).tag(config=True)
    cluster_search_queue = Int(1, min=1,
                               help='search queue size, used if backend="cluster"'
                               ).tag(config=True)
    cluster_summary_interval = Float(60.0,
                                     help='summary interval seconds',
                                     ).tag(config=True)

    dask_search_queue = Int(1, min=1,
                            help='search queue size, used if backend="dask"'
                            ).tag(config=True)
    dask_search_executors = Int(3, min=1,
                                help='search executor number, used if backend="dask"'
                                ).tag(config=True)

    grpc_worker_count = Int(10, min=1,
                            help='grpc worker count'
                            ).tag(config=True)



# -*- coding:utf-8 -*-
"""

"""
import enum

from hypernets.utils import to_repr
from .stateful import Stateful


class OptimizeDirection(enum.Enum):
    Minimize = 'min'
    Maximize = 'max'


class Searcher(Stateful):
    def __init__(self, space_fn, optimize_direction=OptimizeDirection.Minimize, use_meta_learner=True,
                 space_sample_validation_fn=None):
        self.space_fn = space_fn
        self.use_meta_learner = use_meta_learner
        self.optimize_direction = optimize_direction
        self.meta_learner = None
        self.space_sample_validation_fn = space_sample_validation_fn

    def set_meta_learner(self, meta_learner):
        self.meta_learner = meta_learner

    @property
    def parallelizable(self):
        return False

    def sample(self):
        raise NotImplementedError

    def _random_sample(self):
        space_sample = self.space_fn()
        space_sample.random_sample()
        return space_sample

    def _sample_and_check(self, sample_fn):
        counter = 0
        while True:
            space_sample = sample_fn()
            counter += 1
            if counter >= 1000:
                raise ValueError('Unable to take valid sample and exceed the retry limit 1000.')
            if self.space_sample_validation_fn is not None:
                if self.space_sample_validation_fn(space_sample):
                    break
            else:
                break
        return space_sample

    def get_best(self):
        raise NotImplementedError

    def update_result(self, space, result):
        raise NotImplementedError

    def summary(self):
        return 'No Summary'

    def reset(self):
        raise NotImplementedError

    def export(self):
        raise NotImplementedError

    def __repr__(self):
        return to_repr(self)



# -*- coding:utf-8 -*-
"""

"""
from sklearn.metrics import get_scorer

from hypernets.discriminators import make_discriminator
from hypernets.experiment import CompeteExperiment
from hypernets.experiment.cfg import ExperimentCfg as cfg
from hypernets.model import HyperModel
from hypernets.searchers import make_searcher, PlaybackSearcher
from hypernets.tabular import dask_ex as dex
from hypernets.tabular.cache import clear as _clear_cache
from hypernets.tabular.metrics import metric_to_scoring
from hypernets.utils import const, load_data, infer_task_type, hash_data, logging, isnotebook, load_module, DocLens

logger = logging.get_logger(__name__)


def make_experiment(hyper_model_cls,
                    train_data,
                    target=None,
                    eval_data=None,
                    test_data=None,
                    task=None,
                    id=None,
                    callbacks=None,
                    searcher=None,
                    searcher_options=None,
                    search_space=None,
                    search_callbacks=None,
                    early_stopping_rounds=10,
                    early_stopping_time_limit=3600,
                    early_stopping_reward=None,
                    reward_metric=None,
                    optimize_direction=None,
                    hyper_model_options=None,
                    clear_cache=None,
                    discriminator=None,
                    log_level=None,
                    **kwargs):
    """
    Utility to make CompeteExperiment instance with specified settings.

    Parameters
    ----------
    hyper_model_cls: subclass of HyperModel
        Subclass of HyperModel to run trials within the experiment.
    train_data : str, Pandas or Dask DataFrame
        Feature data for training with target column.
        For str, it's should be the data path in file system,
        we'll detect data format from this path (only .csv and .parquet are supported now) and read it.
    target : str, optional
        Target feature name for training, which must be one of the drain_data columns, default is 'y'.
    eval_data : str, Pandas or Dask DataFrame, optional
        Feature data for evaluation with target column.
        For str, it's should be the data path in file system,
        we'll detect data format from this path (only .csv and .parquet are supported now) and read it.
    test_data : str, Pandas or Dask DataFrame, optional
        Feature data for testing without target column.
        For str, it's should be the data path in file system,
        we'll detect data format from this path (only .csv and .parquet are supported now) and read it.
    task : str or None, (default=None)
        Task type(*binary*, *multiclass* or *regression*).
        If None, inference the type of task automatically
    id : str or None, (default=None)
        The experiment id.
    callbacks: list of ExperimentCallback, optional
        ExperimentCallback list.
    searcher : str, searcher class, search object, optional
        The hypernets Searcher instance to explore search space, default is EvolutionSearcher instance.
        For str, should be one of 'evolution', 'mcts', 'random'.
        For class, should be one of EvolutionSearcher, MCTSSearcher, RandomSearcher, or subclass of hypernets Searcher.
        For other, should be instance of hypernets Searcher.
    searcher_options: dict, optional, default is None
        The options to create searcher, is used if searcher is str.
    search_space : callable, optional
        Used to initialize searcher instance (if searcher is None, str or class).
    search_callbacks
        Hypernets search callbacks, used to initialize searcher instance (if searcher is None, str or class).
        If log_level >= WARNNING, default is EarlyStoppingCallback only.
        If log_level < WARNNING, defalult is EarlyStoppingCallback plus SummaryCallback.
    early_stopping_rounds :　int, optional
        Setting of EarlyStoppingCallback, is used if EarlyStoppingCallback instance not found from search_callbacks.
        Set zero or None to disable it, default is 10.
    early_stopping_time_limit : int, optional
        Setting of EarlyStoppingCallback, is used if EarlyStoppingCallback instance not found from search_callbacks.
        Set zero or None to disable it, default is 3600 seconds.
    early_stopping_reward : float, optional
        Setting of EarlyStoppingCallback, is used if EarlyStoppingCallback instance not found from search_callbacks.
        Set zero or None to disable it, default is None.
    reward_metric : str, callable, optional, (default 'accuracy' for binary/multicalss task, 'rmse' for regression task)
        Hypernets search reward metric name or callable. Possible values:
            - accuracy
            - auc
            - f1
            - logloss
            - mse
            - mae
            - msle
            - precision
            - rmse
            - r2
            - recall
    optimize_direction : str, optional
        Hypernets search reward metric direction, default is detected from reward_metric.
    discriminator : instance of hypernets.discriminator.BaseDiscriminator, optional
        Discriminator is used to determine whether to continue training
    hyper_model_options: dict, optional
        Options to initlize HyperModel except *reward_metric*, *task*, *callbacks*, *discriminator*.
    clear_cache: bool, optional, (default False)
    log_level : int, str, or None, (default=None),
        Level of logging, possible values:
            -logging.CRITICAL
            -logging.FATAL
            -logging.ERROR
            -logging.WARNING
            -logging.WARN
            -logging.INFO
            -logging.DEBUG
            -logging.NOTSET
    kwargs:
        Parameters to initialize experiment instance, refrence CompeteExperiment for more details.
    Returns
    -------
    Runnable experiment object

    Examples:
    -------
    Create experiment with example PlainModel and PlainSearchSpace:
    >>> from hypernets.examples.plain_model import PlainModel, PlainSearchSpace
    >>> from hypernets.experiment import make_experiment
    >>> from hypernets.tabular.datasets import dsutils
    >>> df = dsutils.load_blood()
    >>> experiment = make_experiment(PlainModel, df, target='Class', search_space=PlainSearchSpace())
    >>> estimator = experiment.run()

    """
    assert callable(hyper_model_cls) and hasattr(hyper_model_cls, '__name__')
    assert train_data is not None, 'train data is required.'
    if not issubclass(hyper_model_cls, HyperModel):
        logger.warning(f'{hyper_model_cls.__name__} isn\'t subclass of HyperModel.')

    kwargs = kwargs.copy()

    if log_level is None:
        log_level = logging.WARN
    logging.set_level(log_level)

    def find_target(df):
        columns = df.columns.to_list()
        for col in columns:
            if col.lower() in cfg.experiment_default_target_set:
                return col
        raise ValueError(f'Not found one of {cfg.experiment_default_target_set} from your data,'
                         f' implicit target must be specified.')

    def default_searcher(cls, options):
        assert search_space is not None, '"search_space" should be specified when "searcher" is None or str.'
        assert optimize_direction in {'max', 'min'}
        if options is None:
            options = {}
        options['optimize_direction'] = optimize_direction
        s = make_searcher(cls, search_space, **options)

        return s

    def to_search_object(sch):
        from hypernets.core.searcher import Searcher as SearcherSpec
        from hypernets.searchers import EvolutionSearcher

        if sch is None:
            sch = default_searcher(EvolutionSearcher, searcher_options)
        elif isinstance(sch, (type, str)):
            sch = default_searcher(sch, searcher_options)
        elif not isinstance(sch, SearcherSpec):
            logger.warning(f'Unrecognized searcher "{sch}".')

        return sch

    def default_experiment_callbacks():
        cbs = cfg.experiment_callbacks_notebook if isnotebook() else cfg.experiment_callbacks_console
        cbs = [load_module(cb)() if isinstance(cb, str) else cb for cb in cbs]
        return cbs

    def default_search_callbacks():
        cbs = cfg.hyper_model_callbacks_notebook if isnotebook() else cfg.hyper_model_callbacks_console
        cbs = [load_module(cb)() if isinstance(cb, str) else cb for cb in cbs]
        return cbs

    def append_early_stopping_callbacks(cbs):
        from hypernets.core.callbacks import EarlyStoppingCallback

        assert isinstance(cbs, (tuple, list))
        if any([isinstance(cb, EarlyStoppingCallback) for cb in cbs]):
            return cbs

        op = optimize_direction if optimize_direction is not None \
            else 'max' if scorer._sign > 0 else 'min'
        es = EarlyStoppingCallback(early_stopping_rounds, op,
                                   time_limit=early_stopping_time_limit,
                                   expected_reward=early_stopping_reward)

        return [es] + cbs

    X_train, X_eval, X_test = [load_data(data) if data is not None else None
                               for data in (train_data, eval_data, test_data)]

    X_train, X_eval, X_test = [dex.reset_index(x) if dex.is_dask_dataframe(x) else x
                               for x in (X_train, X_eval, X_test)]

    if target is None:
        target = find_target(X_train)

    y_train = X_train.pop(target)
    y_eval = X_eval.pop(target) if X_eval is not None else None

    if task is None:
        dc_nan_chars = kwargs.get('data_cleaner_args', {}).get('nan_chars')
        if isinstance(dc_nan_chars, str):
            dc_nan_chars = [dc_nan_chars]
        task, _ = infer_task_type(y_train, excludes=dc_nan_chars if dc_nan_chars is not None else None)

    if reward_metric is None:
        reward_metric = 'rmse' if task == const.TASK_REGRESSION else 'accuracy'
        logger.info(f'no reward metric specified, use "{reward_metric}" for {task} task by default.')

    if kwargs.get('scorer') is None:
        scorer = metric_to_scoring(reward_metric, task=task, pos_label=kwargs.get('pos_label'))
    else:
        scorer = kwargs.pop('scorer')

    if isinstance(scorer, str):
        scorer = get_scorer(scorer)

    if optimize_direction is None or len(optimize_direction) == 0:
        optimize_direction = 'max' if scorer._sign > 0 else 'min'

    searcher = to_search_object(searcher)

    if cfg.experiment_auto_down_sample_enabled and not isinstance(searcher, PlaybackSearcher) \
            and 'down_sample_search' not in kwargs.keys():
        train_data_shape = dex.compute(X_train.shape)[0] if dex.is_dask_object(X_train) else X_train.shape
        if train_data_shape[0] > cfg.experiment_auto_down_sample_rows_threshold:
            kwargs['down_sample_search'] = True

    if search_callbacks is None:
        search_callbacks = default_search_callbacks()
    search_callbacks = append_early_stopping_callbacks(search_callbacks)

    if callbacks is None:
        callbacks = default_experiment_callbacks()

    if discriminator is None and cfg.experiment_discriminator is not None and len(cfg.experiment_discriminator) > 0:
        discriminator = make_discriminator(cfg.experiment_discriminator,
                                           optimize_direction=optimize_direction,
                                           **(cfg.experiment_discriminator_options or {}))

    if id is None:
        id = hash_data(dict(X_train=X_train, y_train=y_train, X_test=X_test, X_eval=X_eval, y_eval=y_eval,
                            eval_size=kwargs.get('eval_size'), target=target, task=task))
        id = f'{hyper_model_cls.__name__}_{id}'

    if hyper_model_options is None:
        hyper_model_options = {}
    hm = hyper_model_cls(searcher, reward_metric=reward_metric, task=task, callbacks=search_callbacks,
                         discriminator=discriminator, **hyper_model_options)

    experiment = CompeteExperiment(hm, X_train, y_train, X_eval=X_eval, y_eval=y_eval, X_test=X_test,
                                   task=task, id=id, callbacks=callbacks, scorer=scorer, **kwargs)

    if clear_cache:
        _clear_cache()

    if logger.is_info_enabled():
        train_shape, test_shape, eval_shape = \
            dex.compute(X_train.shape,
                        X_test.shape if X_test is not None else None,
                        X_eval.shape if X_eval is not None else None,
                        traverse=True)
        logger.info(f'make_experiment with train data:{train_shape}, '
                    f'test data:{test_shape}, eval data:{eval_shape}, target:{target}')

    return experiment


def _merge_doc():
    my_doc = DocLens(make_experiment.__doc__)
    exp_doc = DocLens(CompeteExperiment.__init__.__doc__)
    excluded = ['hyper_model', 'X_train', 'y_train', 'X_eval', 'y_eval', 'X_test']
    params = my_doc.merge_parameters(exp_doc, exclude=excluded)
    for k in ['clear_cache', 'log_level']:
        params.move_to_end(k)
    params.pop('kwargs')
    my_doc.parameters = params

    make_experiment.__doc__ = my_doc.render()


_merge_doc()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from hypernets.dispatchers.predict.grpc.proto import predict_pb2 as hypernets_dot_dispatchers_dot_predict_dot_grpc_dot_proto_dot_predict__pb2


class PredictServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.predict = channel.unary_unary(
        '/hypernets.dispatchers.predict.grpc.proto.PredictService/predict',
        request_serializer=hypernets_dot_dispatchers_dot_predict_dot_grpc_dot_proto_dot_predict__pb2.PredictRequest.SerializeToString,
        response_deserializer=hypernets_dot_dispatchers_dot_predict_dot_grpc_dot_proto_dot_predict__pb2.PredictResponse.FromString,
        )


class PredictServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def predict(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_PredictServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'predict': grpc.unary_unary_rpc_method_handler(
          servicer.predict,
          request_deserializer=hypernets_dot_dispatchers_dot_predict_dot_grpc_dot_proto_dot_predict__pb2.PredictRequest.FromString,
          response_serializer=hypernets_dot_dispatchers_dot_predict_dot_grpc_dot_proto_dot_predict__pb2.PredictResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'hypernets.dispatchers.predict.grpc.proto.PredictService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



import math
import os

import dask.dataframe as dd
import numpy as np
import pandas as pd
import psutil

from hypernets.tabular import dask_ex as dex
from hypernets.tabular.datasets import dsutils
from hypernets.utils import const


def _startup_dask(overload):
    from dask.distributed import LocalCluster, Client

    if os.environ.get('DASK_SCHEDULER_ADDRESS') is not None:
        # use dask default settings
        client = Client()
    else:
        # start local cluster
        cores = psutil.cpu_count()
        workers = math.ceil(cores / 3)
        if workers > 1:
            if overload <= 0:
                overload = 1.0
            mem_total = psutil.virtual_memory().available / (1024 ** 3)  # GB
            mem_per_worker = math.ceil(mem_total / workers * overload)
            if mem_per_worker > mem_total:
                mem_per_worker = mem_total
            cluster = LocalCluster(processes=True, n_workers=workers, threads_per_worker=4,
                                   memory_limit=f'{mem_per_worker}GB')
        else:
            cluster = LocalCluster(processes=False)

        client = Client(cluster)
    return client


def setup_dask(cls):
    try:
        from dask.distributed import default_client
        client = default_client()
    except:
        client = _startup_dask(2.0)
    print('Dask Client:', client)

    if cls is not None:
        setattr(cls, 'dask_client_', client)


class Test_DaskCustomizedTransformer:
    @classmethod
    def setup_class(cls):
        setup_dask(cls)

        cls.bank_data = dd.from_pandas(dsutils.load_bank(), npartitions=2)
        cls.movie_lens = dd.from_pandas(dsutils.load_movielens(), npartitions=2)

    def test_lgbm_leaves_encoder_binary(self):
        X = self.bank_data.copy()
        y = X.pop('y')
        cats = X.select_dtypes(['int', 'int64', ]).columns.to_list()
        continous = X.select_dtypes(['float', 'float64']).columns.to_list()
        X = X[cats + continous]
        n_estimators = 50
        X_shape = dex.compute(X.shape)[0]

        # with fit_transform
        t = dex.LgbmLeavesEncoder(cat_vars=cats, cont_vars=continous, task=const.TASK_BINARY,
                                  n_estimators=n_estimators)
        Xt = t.fit_transform(X.copy(), y)

        # check attributes
        attributes = {'lgbm', 'cat_vars', 'cont_vars', 'new_columns', 'task', 'lgbm_params'}
        assert len(attributes - set(dir(t))) == 0
        assert all([hasattr(t, a) for a in attributes])
        assert all([getattr(t, a, None) is not None for a in attributes])

        # check result
        Xt = Xt.compute()
        assert getattr(t.lgbm, 'n_estimators', 0) > 0
        assert len(Xt.columns) == len(cats) + len(continous) + t.lgbm.n_estimators
        assert X_shape[0] == Xt.shape[0]

        # with fit + transform
        t2 = dex.LgbmLeavesEncoder(cat_vars=cats, cont_vars=continous, task=const.TASK_BINARY,
                                   n_estimators=n_estimators)
        Xt = t2.fit(X.copy(), y).transform(X.copy())
        Xt = Xt.compute()
        assert getattr(t2.lgbm, 'n_estimators', 0) > 0
        assert len(Xt.columns) == len(cats) + len(continous) + t2.lgbm.n_estimators
        assert X_shape[0] == Xt.shape[0]

    def test_cat_encoder(self):
        X = self.bank_data.copy()
        y = X.pop('y')
        cats = X.select_dtypes(['int', 'int64', ]).columns.to_list()
        t = dex.CategorizeEncoder(columns=cats)
        Xt = t.fit_transform(X.copy(), y)
        assert isinstance(Xt, dd.DataFrame)

        Xt = Xt.compute()
        assert len(t.new_columns) == len(cats)
        assert len(Xt.columns) == len(X.columns) + len(t.new_columns)

    def run_bins_discretizer(self, strategy, dtype):
        X = self.bank_data.copy()

        encoder = dex.MultiKBinsDiscretizer(['age', 'id'], strategy=strategy, dtype=dtype)
        X = encoder.fit_transform(X)
        assert set(encoder.bin_edges_.keys()) == {'age', 'id'}

        assert isinstance(X, dd.DataFrame)
        X = X.compute()

        for col, new_name, c_bins in encoder.new_columns:
            assert X[new_name].nunique() <= c_bins

            col_values = X[new_name].values
            assert col_values.min() == 0 and col_values.max() == c_bins - 1

            if dtype is not None and dtype.startswith('float'):
                assert str(X[new_name].dtype).startswith('float')
            else:
                assert str(X[new_name].dtype).startswith('int')

    def test_bins_discretizer_uniform_float64(self):
        self.run_bins_discretizer('uniform', 'float64')

    def test_bins_discretizer_quantile_float64(self):
        self.run_bins_discretizer('quantile', 'float64')

    def test_bins_discretizer_uniform_int(self):
        self.run_bins_discretizer('uniform', 'int')

    def test_bins_discretizer_quantile_int(self):
        self.run_bins_discretizer('uniform', None)

    def test_varlen_encoder_with_customized_data(self):
        from io import StringIO
        data = '''
        col_foo
        a|b|c|x
        a|b
        b|b|x
        '''.replace(' ', '')

        df = pd.read_csv(StringIO(data))
        result = pd.DataFrame({'col_foo': [
            [1, 2, 3, 4],
            [1, 2, 0, 0],
            [2, 2, 4, 0]
        ]})

        multi_encoder = dex.MultiVarLenFeatureEncoder([('col_foo', '|')])
        result_df = multi_encoder.fit_transform(df.copy())
        print(result_df)
        assert all(result_df.values == result.values)

        ddf = dd.from_pandas(df.copy(), npartitions=1)
        d_multi_encoder = dex.MultiVarLenFeatureEncoder([('col_foo', '|')])
        d_result_df = d_multi_encoder.fit_transform(ddf)
        assert isinstance(d_result_df, dd.DataFrame)
        d_result_df = d_result_df.compute()

        print(d_result_df)
        assert all(d_result_df.values == result.values)

    def test_dataframe_wrapper(self):
        X = self.bank_data.copy()

        cats = X.select_dtypes(['object', ]).columns.to_list()
        continous = X.select_dtypes(['float', 'float64', 'int', 'int64']).columns.to_list()
        transformers = [('cats',
                         dex.SimpleImputer(missing_values=np.nan, strategy='constant', fill_value=''),
                         cats
                         ),
                        ('conts',
                         dex.SimpleImputer(missing_values=np.nan, strategy='mean'),
                         continous
                         )]
        ct = dex.ColumnTransformer(transformers=transformers)
        dfw = dex.DataFrameWrapper(ct, cats + continous)
        X = dfw.fit_transform(X[cats + continous])

        assert dex.is_dask_dataframe(X)
        print(X.dtypes)


if __name__ == '__main__':
    Test_DaskCustomizedTransformer.setup_class()
    c = Test_DaskCustomizedTransformer()
    c.test_cat_encoder()



# -*- coding:utf-8 -*-
"""

"""
import numpy as np

from hypernets.core.ops import Identity
from hypernets.core.search_space import HyperSpace, Int, Real, Choice, Bool
from hypernets.core.searcher import OptimizeDirection
from hypernets.searchers.evolution_searcher import Population, EvolutionSearcher


def get_space():
    space = HyperSpace()
    with space.as_default():
        p1 = Int(1, 100)
        p2 = Choice(['a', 'b'])
        p3 = Bool()
        p4 = Real(0.0, 1.0)
        id1 = Identity(p1=p1)
        id2 = Identity(p2=p2)(id1)
        id3 = Identity(p3=p3)(id2)
        id4 = Identity(p4=p4)(id3)
    return space


class Test_Evolution():

    def test_population(self):
        population = Population(optimize_direction=OptimizeDirection.Maximize, random_state=np.random.RandomState(9527))
        population.append('a', 0)
        population.append('b', 1)
        population.append('c', 2)
        population.append('d', 3)
        population.append('e', 4)
        population.append('f', 5)
        population.append('g', 6)
        population.append('h', 7)
        population.append('i', 8)
        population.append('i', 9)

        b1 = population.sample_best(25)
        assert b1.reward == 8

        population = Population(optimize_direction=OptimizeDirection.Minimize, random_state=np.random.RandomState(9527))
        population.append('a', 0)
        population.append('b', 1)
        population.append('c', 2)
        population.append('d', 3)
        population.append('e', 4)
        population.append('f', 5)
        population.append('g', 6)
        population.append('h', 7)
        population.append('i', 8)
        population.append('i', 9)

        b2 = population.sample_best(25)
        assert b2.reward == 0

    def test_eliminate(self):
        population = Population(optimize_direction=OptimizeDirection.Maximize)
        population.append('a', 4)
        population.append('b', 3)
        population.append('c', 2)
        population.append('d', 1)
        population.append('e', 0)
        population.append('f', 5)
        population.append('g', 6)
        population.append('h', 7)
        population.append('i', 8)
        population.append('j', 9)

        eliminates = population.eliminate(2, regularized=True)
        assert eliminates[0].space_sample == 'a' and eliminates[1].space_sample == 'b'

        eliminates = population.eliminate(2, regularized=False)
        assert eliminates[0].space_sample == 'e' and eliminates[1].space_sample == 'd'

    def test_mutate(self):
        def get_space():
            space = HyperSpace()
            with space.as_default():
                id1 = Identity(p1=Int(0, 10), p2=Choice(['a', 'b']))
                id2 = Identity(p3=Real(0., 1.), p4=Bool())(id1)
            return space

        # population = Population(optimize_direction=OptimizeDirection.Maximize)
        population = Population(optimize_direction='max')

        space1 = get_space()
        space1.random_sample()
        assert space1.all_assigned

        space2 = get_space()
        assert not space2.all_assigned
        new_space = population.mutate(space1, space2)

        pv1 = list(space1.get_assigned_param_values().values())
        pv2 = list(space2.get_assigned_param_values().values())

        assert space2.all_assigned
        assert new_space.all_assigned
        assert np.sum([v1 != v2 for v1, v2 in zip(pv1, pv2)]) == 1

    def test_set_random_state(self):
        from hypernets.core import set_random_state
        set_random_state(9527)

        searcher = EvolutionSearcher(get_space, 5, 3, regularized=False, optimize_direction=OptimizeDirection.Maximize)
        vectors = []
        for i in range(1, 10):
            vectors.append(searcher.sample().vectors)
        assert vectors == [[98, 0, 0, 0.96], [9, 0, 0, 0.93], [60, 0, 1, 0.24], [54, 0, 1, 0.7],
                           [25, 0, 1, 0.73], [67, 1, 1, 0.43], [57, 1, 1, 0.05], [49, 0, 0, 0.71], [71, 1, 1, 0.49]]

        set_random_state(None)
        searcher = EvolutionSearcher(get_space, 5, 3, regularized=False, optimize_direction=OptimizeDirection.Maximize)
        vectors = []
        for i in range(1, 10):
            vectors.append(searcher.sample().vectors)
        assert vectors != [[98, 0, 0, 0.96], [9, 0, 0, 0.93], [60, 0, 1, 0.24], [54, 0, 1, 0.7],
                           [25, 0, 1, 0.73], [67, 1, 1, 0.43], [57, 1, 1, 0.05], [49, 0, 0, 0.71], [71, 1, 1, 0.49]]

        set_random_state(9527)
        searcher = EvolutionSearcher(get_space, 5, 3, regularized=False, optimize_direction=OptimizeDirection.Maximize)
        vectors = []
        for i in range(1, 10):
            vectors.append(searcher.sample().vectors)
        assert vectors == [[98, 0, 0, 0.96], [9, 0, 0, 0.93], [60, 0, 1, 0.24], [54, 0, 1, 0.7],
                           [25, 0, 1, 0.73], [67, 1, 1, 0.43], [57, 1, 1, 0.05], [49, 0, 0, 0.71], [71, 1, 1, 0.49]]

        set_random_state(1)
        searcher = EvolutionSearcher(get_space, 5, 3, regularized=False, optimize_direction=OptimizeDirection.Maximize)
        vectors = []
        for i in range(1, 10):
            vectors.append(searcher.sample().vectors)
        assert vectors == [[38, 1, 0, 0.93], [10, 1, 1, 0.15], [17, 1, 0, 0.39], [7, 1, 0, 0.85], [19, 0, 1, 0.44],
                           [29, 1, 0, 0.67], [88, 1, 1, 0.43], [95, 0, 0, 0.8], [10, 1, 1, 0.09]]

        set_random_state(None)

    # def test_searcher_with_hp(self):
    #     def get_space():
    #         space = HyperSpace()
    #         with space.as_default():
    #             in1 = Input(shape=(10,))
    #             in2 = Input(shape=(20,))
    #             in3 = Input(shape=(1,))
    #             concat = Concatenate()([in1, in2, in3])
    #             dense1 = Dense(10, activation=Choice(['relu', 'tanh', None]), use_bias=Bool())(concat)
    #             bn1 = BatchNormalization()(dense1)
    #             dropout1 = Dropout(Choice([0.3, 0.4, 0.5]))(bn1)
    #             output = Dense(2, activation='softmax', use_bias=True)(dropout1)
    #         return space
    #
    #     rs = EvolutionSearcher(get_space, 5, 3, regularized=False, optimize_direction=OptimizeDirection.Maximize)
    #     hk = HyperKeras(rs, optimizer='adam', loss='sparse_categorical_crossentropy', metrics=['accuracy'],
    #                     callbacks=[SummaryCallback()])
    #
    #     x1 = np.random.randint(0, 10000, size=(100, 10))
    #     x2 = np.random.randint(0, 100, size=(100, 20))
    #     x3 = np.random.normal(1.0, 100.0, size=(100))
    #     y = np.random.randint(0, 2, size=(100), dtype='int')
    #     x = [x1, x2, x3]
    #
    #     hk.search(x, y, x, y, max_trials=10)
    #     assert hk.get_best_trial()
    #     best_trial = hk.get_best_trial()
    #
    #     estimator = hk.final_train(best_trial.space_sample, x, y)
    #     score = estimator.predict(x)
    #     result = estimator.evaluate(x, y)
    #     assert len(score) == 100
    #     assert result



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

import io

import numpy as np
import pandas as pd
from dask import dataframe as dd
from numpy import dtype

from hypernets.tabular.data_cleaner import DataCleaner

csv_str = '''x1_int_nanchar,x2_all_nan,x3_const_str,x4_const_int,x5_dup_1,x6_dup_2,x7_dup_f1,x8_dup_f2,x9_f,x10,y
1.0,,const,5,dup,dup,0.1,0.1,1.23,\\N,1
2.2,,const,5,dupa,dupa,0.111,0.111,4.4,\\N,1
\\N,,const,5,dupb,dupb,0.12323,0.12323,1.233,\\N,1
4.,,const,5,dupc,dupc,0.14334,0.14334,4534434.2,\\N,0
5,,const,5,dupd,dupd,0.144,0.144,2302.2,\\N,0
6,,const,5,dupe,dupe,0.155,0.155,34334.1,\\N,\\N
'''


class Test_DataCleaner():
    def test_basic(self):
        pdf = pd.read_csv(io.StringIO(csv_str))
        ddf = dd.from_pandas(pdf, npartitions=2)

        for df in (pdf, ddf):
            # assert df.shape == (6, 11)
            assert df.shape[1] == 11
            assert list(df.dtypes.values) == [dtype('O'), dtype('float64'), dtype('O'), dtype('int64'), dtype('O'),
                                              dtype('O'), dtype('float64'), dtype('float64'), dtype('float64'),
                                              dtype('O'),
                                              dtype('O')]

            y = df.pop('y')
            cleaner = DataCleaner(nan_chars='\\N',
                                  correct_object_dtype=True,
                                  drop_constant_columns=True,
                                  drop_label_nan_rows=True,
                                  drop_duplicated_columns=True,
                                  drop_idness_columns=False,
                                  replace_inf_values=np.nan
                                  )

            x_t, y_t = cleaner.fit_transform(df, y)
            if isinstance(df, dd.DataFrame):
                x_t, y_t = x_t.compute(), y_t.compute()
            assert x_t.shape == (5, 4)
            assert y_t.shape == (5,)
            assert x_t.columns.to_list() == ['x1_int_nanchar', 'x5_dup_1', 'x7_dup_f1', 'x9_f']
            assert list(x_t.dtypes.values) == [dtype('float64'), dtype('O'), dtype('float64'), dtype('float64')]
            assert cleaner.df_meta_ == {'float64': ['x1_int_nanchar', 'x7_dup_f1', 'x9_f'], 'object': ['x5_dup_1']}

            cleaner.append_drop_columns(['x9_f'])

            assert cleaner.df_meta_ == {'float64': ['x1_int_nanchar', 'x7_dup_f1'], 'object': ['x5_dup_1']}
            x_t, y_t = cleaner.transform(df, y)
            if isinstance(df, dd.DataFrame):
                x_t, y_t = x_t.compute(), y_t.compute()
            assert x_t.shape == (5, 3)
            assert y_t.shape == (5,)
            assert x_t.columns.to_list() == ['x1_int_nanchar', 'x5_dup_1', 'x7_dup_f1']
            assert list(x_t.dtypes.values) == [dtype('float64'), dtype('O'), dtype('float64')]

            cleaner = DataCleaner(nan_chars='\\N',
                                  correct_object_dtype=True,
                                  drop_constant_columns=True,
                                  drop_label_nan_rows=True,
                                  drop_duplicated_columns=False,
                                  drop_idness_columns=False,
                                  replace_inf_values=np.nan
                                  )

            x_t, y_t = cleaner.fit_transform(df, y)
            if isinstance(df, dd.DataFrame):
                x_t, y_t = x_t.compute(), y_t.compute()
            assert x_t.shape == (5, 6)
            assert y_t.shape == (5,)
            assert x_t.columns.to_list() == ['x1_int_nanchar', 'x5_dup_1', 'x6_dup_2', 'x7_dup_f1', 'x8_dup_f2', 'x9_f']
            assert list(x_t.dtypes.values) == [dtype('float64'), dtype('O'), dtype('O'), dtype('float64'),
                                               dtype('float64'),
                                               dtype('float64')]
            assert cleaner.df_meta_ == {'float64': ['x1_int_nanchar', 'x7_dup_f1', 'x8_dup_f2', 'x9_f'],
                                        'object': ['x5_dup_1', 'x6_dup_2']}

            cleaner = DataCleaner(nan_chars='\\N',
                                  correct_object_dtype=True,
                                  drop_constant_columns=True,
                                  drop_label_nan_rows=False,
                                  drop_duplicated_columns=False,
                                  drop_idness_columns=False,
                                  replace_inf_values=np.nan
                                  )

            x_t, y_t = cleaner.fit_transform(df, y)
            if isinstance(df, dd.DataFrame):
                x_t, y_t = x_t.compute(), y_t.compute()
            assert x_t.shape == (6, 6)
            assert y_t.shape == (6,)

            cleaner = DataCleaner(nan_chars='\\N',
                                  correct_object_dtype=False,
                                  drop_constant_columns=True,
                                  drop_label_nan_rows=False,
                                  drop_duplicated_columns=False,
                                  drop_idness_columns=False,
                                  replace_inf_values=np.nan
                                  )

            x_t, y_t = cleaner.fit_transform(df, y)
            if isinstance(df, dd.DataFrame):
                x_t, y_t = x_t.compute(), y_t.compute()
            assert x_t.shape == (6, 6)
            assert y_t.shape == (6,)
            assert x_t.columns.to_list() == ['x1_int_nanchar', 'x5_dup_1', 'x6_dup_2', 'x7_dup_f1', 'x8_dup_f2', 'x9_f']
            assert list(x_t.dtypes.values) == [dtype('O'), dtype('O'), dtype('O'), dtype('float64'), dtype('float64'),
                                               dtype('float64')]
            assert cleaner.df_meta_ == {'object': ['x1_int_nanchar', 'x5_dup_1', 'x6_dup_2'],
                                        'float64': ['x7_dup_f1', 'x8_dup_f2', 'x9_f']}

            cleaner = DataCleaner(nan_chars='\\N',
                                  correct_object_dtype=False,
                                  drop_constant_columns=False,
                                  drop_label_nan_rows=False,
                                  drop_duplicated_columns=False,
                                  drop_idness_columns=False,
                                  replace_inf_values=np.nan
                                  )

            x_t, y_t = cleaner.fit_transform(df, y)
            if isinstance(df, dd.DataFrame):
                x_t, y_t = x_t.compute(), y_t.compute()
            assert x_t.shape == (6, 10)
            assert y_t.shape == (6,)

            # not drop
            cleaner = DataCleaner(nan_chars='\\N',
                                  correct_object_dtype=False,
                                  drop_constant_columns=True,
                                  drop_label_nan_rows=True,
                                  drop_duplicated_columns=True,
                                  drop_idness_columns=True,
                                  replace_inf_values=np.nan,
                                  reserve_columns=['x4_const_int']
                                  )

            x_t, y_t = cleaner.fit_transform(df, y)
            if isinstance(df, dd.DataFrame):
                x_t, y_t = x_t.compute(), y_t.compute()
            assert 'x4_const_int' in x_t.columns.to_list()
